/*
* COMPONENT_NAME: ldap.clients
*
* ABSTRACT: generic program for LDAP extended operations
*
* ORIGINS: 202,27
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <ldap.h>
#include <ldapssl.h>

#ifndef _WIN32
#include <errno.h>
#endif

/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else 
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( void );

const char* ssl_err2string( int );

#ifndef min
#define min(x, y) ( (x) < (y) ? (x) : (y) )
#endif

#ifndef max
#define max(x, y) ( (x) > (y) ? (x) : (y) )
#endif

#define DEFHOST "localhost"

typedef struct _Option {
    int optionID;
    const char *optionName;
    int optionMinAbbrev;
    int optionNumArg;
}
Option;

/* Replication Error Log LDAP Extended Operation Function prototypes: */
void replErrLogUsage();
void decodeBerReplErrLog( const char *response_OID, struct berval *result );
struct berval * createRequestReplErrLog( LDAP_REPL_ERROR_LOG_OPTION option, unsigned long failure_ID );
void replErrorLog( LDAP_REPL_ERROR_LOG_OPTION option, unsigned long failure_ID );
int parseReplErrLogInputs(int argIndex, int argc, char *argv[], Option *list);

int isOption( char *argument, Option *optionList);
int isOptionExtended( char *argument, Option *optionList);
int getAttributeNumber(char *);


/*Global Varibles for extended operations...*/
char *agreementDn = NULL;
char *contextDn = NULL;
char *action = NULL;
char *scope = NULL;
char *skip = NULL;
char *entryDn = NULL;
char *attribute = NULL;
char *attrType = NULL;
char *attrsType = NULL; /* for getAttributesType*/
char **attrs = NULL;
char *logName = NULL;
char *lines = NULL;
char *udn = NULL;
char *uip = NULL;
int linesFrom = 0;
int linesTo = 0;
int timeout = 0;
int nofailIfExists = LDAP_FALSE;
int all = LDAP_FALSE;
int cancel = LDAP_FALSE;
int del = LDAP_FALSE;
int end = LDAP_FALSE;
int iScope = ALL;
int iMatch = LDAP_FALSE;
LDAPControl **opControls = NULL;
int stringLen = 0;

/* variables for general options for connecting to the server */
char *host = (char*)DEFHOST;
int port = 0;
char *dn = NULL;
char *passwd = NULL;
char *proxydn = NULL;

char *keyFile = NULL;
char *keyPwd = NULL;
char *keyName = NULL;
char *mechanism = NULL;
char *charset = NULL;
char *changeId = NULL;   /* The change Id to be skipped */
int version = 0;
int debugLevel = 0;
int silent, rc = LDAP_SUCCESS;
int keys = LDAP_FALSE;
int tls = LDAP_FALSE;
int ssl = LDAP_FALSE;
int use_FIPS_mode = LDAP_FALSE;
int verbose = LDAP_FALSE;

/* Digest-MD5 */
static char *digest_username = NULL;
static char *digest_realm = NULL;

/* Other Global varibales */
int argIndex = 1;
static LDAP *ld = NULL;
static int ldapversion = LDAP_VERSION3;

/* These variables are used for getting 2 integers for "lines" option in "readlog" operation */
int Inputlength = 0;
int i = 0, j = 0;

Option *optVal = NULL;
Option *extendedOptVal = NULL;

enum
{
    NULLVALUE,
    /* general options for connecting to server */
    HOST,
    PORT,
    DN,
    PASSWD,
    TLS,
    PROXY,
    SSL,
    KEYFILE,
    KEYPW,
    KEYNAME,
    MECHANISM,
    CHARSET,
    VERSION,
    VERBOSE,
    LEVEL,
    OPTQUESTION,
    OPTHELP,
    OPTEXTENDEDOP,
    DIGEST_USERNAME,
    DIGEST_REALM,
    FIPS,

    /* "-op" options */
    EXOPTFULLREP,
    EXOPTQUIESCE,
    EXOPTCONTROLREP,
    EXOPTCONTROLQUEUE,
    EXOPTCASCREP,
    EXOPTREADCONFG,
    EXOPTGETLOGSIZE,
    EXOPTREADLOG,
    EXOPTCLEARLOG,
    EXOPTGETUSRTYPE,
    EXOPTSTOPSERVER,
    EXOPTUNBIND,
    EXOPTATTRTYPE,
    EXOPTGROUPEVAL,
    EXOPTREPLTOPOLOGY,
    EXOPTREPLERRLOG,
    EXOPTACCTSTATUS,
    /* "fullRepl" options */
    FULLREPAGGDN,
    FULLREPALL,
    FULLREPCANCEL,
    FULLREPDEL,
    FULLREPNOFAIL,
    /* "quiesce" options */
    QUIESCERC,
    QUIESCEEND,
    /* "unique attribute" options */
    ATTRTYPE,
    /* "controlrepl" options */
    CONTROLREPLACTION,
    CONTROLREPLRA,
    CONTROLREPLRC,
    /* "controlqueue" options */
    CONTROLQUEUESKIP,
    CONTROLQUEUERA,
    /* "cascrepl" options */
    CASCREPLACTION,
    CASCREPLRC,
    CASCREPLTIMEOUT,
    /* "readconfig" options */
    READCONFIGSCOPE,
    /* "getlogsize" option */
    LOGSIZELOG,
    /* "readlog" option */
    READLOGLOG,
    READLOGLINES,
    /* "clearlog" option */
    CLEARLOGLOG,
    /* "Unbind" options*/
    UNBINDDN,
    UNBINDIP,
    UNBINDALL,
    /* "getattributes" options*/
    MATCHES,
    EXOPTGETATTRTYPE,
    /* "groupevalutaion options*/
    ATTRS,
    /* "repltopology" options*/
    REPLTOPOLOGYRC,
    REPLTOPOTIMEOUT,
    REPLTOPOLOGYRA,
    /* "replerrlog" options*/
    REPLERRLOGSHOW,
    REPLERRLOGDELETE,
    REPLERRLOGRETRY,
    REPLERRLOGAGMTDN
};

/* Scope for "readconfig" option */
const char * SINGLE_SCOPE = "single";
const char * ENTRY_SCOPE = "entry";
const char * SUBTREE_SCOPE = "subtree";
const char * ALL_SCOPE = "entire";

/* List for "clearlog" options */
static Option clearLogList[] =
    {
        { CLEARLOGLOG, "log", 1, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/*List for "readlog" options */
static Option readLogList[] =
    {
        { READLOGLOG, "log", 2, 1 },
        { READLOGLINES, "lines", 2, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };


/*List for "getlogsize" options */
static Option logSizeList[] =
    {
        { LOGSIZELOG, "log", 1, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };


/* List for "readconfig" options */
static Option readConfigList[] =
    {
        { READCONFIGSCOPE, "scope", 1, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* List for "cascrepl" options */
static Option cascadeList[] =
    {
        { CASCREPLACTION, "action", 1, 1 },
        { CASCREPLRC , "rc", 1, 1 },
        { CASCREPLTIMEOUT, "timeout", 1, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* List for "controlqueue" options */
static Option controlQueueList[] =
    {
        { CONTROLQUEUESKIP, "skip", 1, 1 },
        { CONTROLQUEUERA, "ra", 1, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };


/* List for "controlrepl" options */
static Option controlReplList[] =
    {
        { CONTROLREPLACTION, "action", 1, 1 },
        { CONTROLREPLRA, "ra", 1, 1 },
        { CONTROLREPLRC, "rc", 1, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };


/* List for "unbind" options */
static Option unbindList[] =
    {
        { UNBINDDN, "dn", 1, 1 },
        { UNBINDIP, "ip", 1, 1 },
        { UNBINDALL, "all", 1, 1 },
        { OPTQUESTION, "?", 1, 0},
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }

    };

/* list for "unique attribute" options */
static Option uniqueAttrList[] =
    {
        { ATTRTYPE, "a", 1, 0 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* list for "getattributes" options */
static Option getAttrList[] =
    {
        { ATTRTYPE, "attrType", 1, 0 },
        { MATCHES, "matches", 1, 0 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* list for "evaluategroups" options */
static Option evalGroupsList[] =
    {
        { DN, "d", 1, 0 },
        { ATTRS, "a", 0, 0 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* List for "repltopology" options */
static Option replTopologyList[] =
    {
        { REPLTOPOLOGYRC, "rc", 1, 1 },
        { REPLTOPOTIMEOUT, "timeout", 1, 1 },
        { REPLTOPOLOGYRA, "ra", 1, 1 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };


/* list for "acctstatus" options */
static Option acctStatusList[] =
    {
        { DN, "d", 1, 0 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* List for "replerrlog" options */
/*    option ID          name    min args */
static Option replErrLogList[] =
    {
        { REPLERRLOGSHOW,   "show",   1, 1 },
        { REPLERRLOGDELETE, "delete", 1, 1 },
        { REPLERRLOGRETRY,  "retry",  2, 1 },
        { REPLERRLOGAGMTDN, "ra",     2, 1 },
        { OPTQUESTION,      "?",      1, 0 },
        { OPTHELP,          "help",   1, 0 },
        { NULLVALUE,         NULL,    0, 0 }
    };

/* List for general options... */

static Option optList[] =
    {
        { HOST, "h", 1, 1 },
        { PORT, "p", 1, 1 },
        { DN, "D", 1, 1 },
        { PASSWD, "w", 1, 1 },
        { PROXY, "y", 1, 1},
        { TLS, "Y", 1, 0 },
        { SSL, "Z", 1, 0 },
        { KEYFILE, "K", 1, 1 },
        { KEYPW, "P", 1, 1 },
        { KEYNAME, "N", 1, 1 },
        { MECHANISM, "m", 1, 1 },
        { CHARSET, "C", 1, 1 },
        { VERSION, "e", 1, 1 },
        { LEVEL, "d", 1, 1 },
        { VERBOSE, "v", 1, 0 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { OPTEXTENDEDOP, "op", 2, 0 },
        { DIGEST_USERNAME, "U", 1, 1 },
        { DIGEST_REALM, "G", 1, 1 },
        { FIPS, "x", 1, 1 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* List for "-op" options */

static Option extendedOptList[] =
    {
        { EXOPTFULLREP, "fullRepl", 1, 0 },
        { EXOPTQUIESCE, "quiesce", 1, 0 },
        { EXOPTREPLERRLOG, "controlreplerr", 12, 0 },
        { EXOPTCONTROLREP, "controlrepl",11, 1 },
        { EXOPTCONTROLQUEUE, "controlqueue", 8, 0 },
        { EXOPTCASCREP, "cascrepl", 2, 1 },
        { EXOPTREADCONFG, "readconfig", 5, 1 },
        { EXOPTGETLOGSIZE, "getlogsize", 4, 0 },
        { EXOPTREADLOG, "readlog", 5, 0 },
        { EXOPTCLEARLOG, "clearlog", 2, 0 },
        { EXOPTGETUSRTYPE, "getusertype", 4, 0 },
        { EXOPTSTOPSERVER, "stopserver", 1, 0 },
        { EXOPTUNBIND, "unbind", 1, 0 },
        { EXOPTATTRTYPE, "uniqueattr", 1, 0 },
        { EXOPTGETATTRTYPE, "getattributes", 4, 0 },
        { EXOPTGROUPEVAL, "evaluategroups", 1, 0 },
        { EXOPTREPLTOPOLOGY, "repltopology", 8, 0 },
        { EXOPTACCTSTATUS, "acctstatus", 1, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };


/* List for "fullRepl" options */
static Option fullReplList[] =
    {
        { FULLREPAGGDN, "ra", 1, 1 },
        { FULLREPALL, "all", 1, 0 },
        { FULLREPCANCEL, "cancel", 1, 0 },
        { FULLREPDEL, "del", 1, 0 },
        { FULLREPNOFAIL, "nofailIfExists", 1, 0 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* list for "quiesce" options */
static Option quiesceList[] =
    {
        { QUIESCERC, "rc", 1, 1 },
        { QUIESCEEND, "end", 1, 0 },
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

/* list for "getusertype" and "stopserver" options */
static Option commonOptionList[] =
    {
        { OPTQUESTION, "?", 1, 0 },
        { OPTHELP, "help", 2, 0 },
        { NULLVALUE, NULL, 0, 0 }
    };

void printOption(Option *o) {
    printf("\nID                  : %d", o->optionID);
    printf("\nOption Name         : %s", o->optionName);
    printf("\nMinimum Abbrev      : %d", o->optionMinAbbrev);
    printf("\nNumber of Arguments : %d", o->optionNumArg);
}

/*Usage functions*/
void
unbindUsage() {
    fprintf(stderr, "\nThe usage for the unbind operation is : ");
    fprintf(stderr, "\n\n1. -op unbind -dn specificDN");
    fprintf(stderr, "\n2. -op unbind -ip specificIP");
    fprintf(stderr, "\n3. -op unbind -dn specificDN -ip specificIP");
    fprintf(stderr, "\n4. -op unbind -all\n\n");
    fprintf(stderr, "\nwhere");
    fprintf(stderr, "\n  specificDN        specifies a bound DN");
    fprintf(stderr, "\n  specificIP        specifies a source IP");
    fprintf(stderr, "\n  all               specifies all connections\n\n");
    exit(0);
}

void uniqueAttrUsage() {
    fprintf(stderr, "\nThe usage for unique attribute operation is : ");
    fprintf(stderr, "\n\n-op uniqueattr -a attributeType");
    fprintf(stderr, "\n\nwhere:");
    fprintf(stderr, "\nattributeType specifies the attribute for which all conflicting"
            "\nvalues will be listed\n");
    exit(0);
}

void getAttrsUsage() {
    fprintf(stderr, "\nThe usage for getattributes operation is : ");
    fprintf(stderr, "\n\n-op getattributes -attrType type -matches bool");
    fprintf(stderr, "\n\nwhere:");
    fprintf(stderr, "\ntype       operational | language_tag | attribute_cache | unique | configuration"
            "\nbool       true | false\n");
    exit(0);
}

/**********************************************************************
 * Function: groupEvalUsage
 * Purpose:  Print out a usage statement for the evaluate groups operation
 * Inputs: None
 * Pre-conditions: None
 * Post-conditions: program exits
 * Outputs: None          
 ***********************************************************************/
void groupEvalUsage() {
    fprintf(stderr, "\nThe usage for evaluategroups operation is : ");
    fprintf(stderr, "\n\n-op evaluategroups -d DN [-a attribute value pairs...]\n");
    fprintf(stderr, "\nwhere:\n");
    fprintf(stderr,   "   attribute value pairs: whitespace-separated list of\n");
    fprintf(stderr,   "                          attribute value pairs.\n");
    fprintf(stderr,   "                          Each attribute value pair is\n");
    fprintf(stderr,   "                          in the attr=value format.\n");
    exit(0);
}

void clearlogUsage() {
    fprintf(stderr, "\nThe usage for clearlog operation is : ");
    fprintf(stderr, "\n\n-op clearlog -log fileName");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n fileName  audit | bulkload | cli |  specifies the log file to be cleared"
            "\n           slapd | ibmdiradm | adminAudit |"
            "\n           debug | LostAndFound | config\n");
    exit(0);
}

void readlogUsage() {
    fprintf(stderr, "\nThe usage for readlog operation is :");
    fprintf(stderr, "\n\n-op readlog -log fileName -lines value");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n fileName  audit | bulkload | cli |  specifies the log file to be queried"
            "\n           slapd | ibmdiradm | adminAudit |"
            "\n           debug | LostAndFound | config\n");
    fprintf(stderr, "\n value     <first> <last> |  read the specified range of lines"
            "\n           all               read all lines\n");
    exit(0);
}

void getlogsizeUsage() {
    fprintf(stderr, "\nThe usage for getlogsize operation is : ");
    fprintf(stderr, "\n\n-op getlogsize -log fileName");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n fileName  audit | bulkload | cli |  specifies the log file to be queried"
            "\n           slapd | ibmdiradm | adminAudit |"
            "\n           debug | LostAndFound | config\n");
    exit(0);

}

void readconfigUsage() {
    fprintf(stderr, "\nThe usage for readconfig operation is : ");
    fprintf(stderr, "\n\n-op readconfig -scope scopeValue");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n scopeValue  entire                         read the entire file");
    fprintf(stderr, "\n             single  <entry DN> <attribute> read the single \'entry\' and"
            "\n                                            \'attribute\' specified");
    fprintf(stderr, "\n             entry   <entry DN>             read the single \'entry\'");
    fprintf(stderr, "\n             subtree <entry DN>             read the \'entry\' specified"
            "\n                                            and entire subtree under it.\n");
    exit(0);
}

void cascreplUsage() {
    fprintf(stderr, "\nThe usage for cascaded replication operation is : ");
    fprintf(stderr, "\n\n-op cascrepl -action actionValue -rc contextDn [options]");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n actionValue  quiesce     quiesce the context"
            "\n              unquiesce   unquiesce the context"
            "\n              replnow     start immediate replication"
            "\n              wait        wait for pending changes to be replicated"
            "\n contextDn                specifies the root of the subtree");
    fprintf(stderr, "\noptions:");
    fprintf(stderr, "\n -timeout secs            timeout period in seconds\n");
    exit(0);
}

void controlqueueUsage() {
    fprintf(stderr, "\nThe usage for control replication queue operation is : ");
    fprintf(stderr, "\n\n-op controlqueue -skip skipValue -ra agreementDn");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n skipValue   all          skip all pending changes for this agreement");
    fprintf(stderr, "\n             change-id    skip the specified change ");
    fprintf(stderr, "\n agreementDn              DN of the replication agreement\n");
    exit(0);
}

void controlreplUsage() {
    fprintf(stderr, "\nThe usage for control replication operation is : ");
    fprintf(stderr, "\n\n-op controlrepl -action actionValue -ra agreementDn");
    fprintf(stderr, "\nor");
    fprintf(stderr, "\n-op controlrepl -action actionValue -rc contextDn");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n actionValue  suspend    suspend replication"
            "\n              resume     resume replication"
            "\n              replnow    start immediate replication");
    fprintf(stderr, "\n contextDn               specifies the root of the replication context."
            "\n                         The action will be performed for all agreements"
            "\n                         for this context");
    fprintf(stderr, "\n agreementDn             specifies the replication agreement. The action"
            "\n                         will be performed for the specified agreement.\n");
    exit(0);
}

void quiesceUsage() {
    fprintf(stderr, "\nThe usage for quiesce operation is : ");
    fprintf(stderr, "\n\n-op quiesce -rc contextDn [options]");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n contextDn   DN of the replication context (subtree) to be quiesced"
            "\n             or unquiesced");
    fprintf(stderr, "\noptions:");
    fprintf(stderr, "\n -end        unquiesce the subtree\n");
    exit(0);
}

void replTopologyUsage() {
    fprintf(stderr, "\nThe usage for repltopology operation is : ");
    fprintf(stderr, "\n\n-op repltopology -rc contextDn [options]");
    fprintf(stderr, "\n\nwhere:");
    fprintf(stderr, "\n contextDn   specifies the root of the subtree");
    fprintf(stderr, "\noptions:");
    fprintf(stderr, "\n -timeout secs           timeout period in seconds");
    fprintf(stderr, "\n -ra agreementDn         specifies the replication agreement. Entries"
            "\n                         will be propagate to the specified agreement.\n");

    exit(0);
}

void replErrLogUsage() {
    fprintf(stderr, "\nThe usage for controlreplerr operation is : ");
    fprintf(stderr, "\n\n-op controlreplerr -show failure_ID -ra agreementDn");
    fprintf(stderr, "\n\nwhere:");
    fprintf(stderr, "\n     failure_ID  specifies the ID of the target replication update failure.");
    fprintf(stderr, "\n     agreementDn specifies the DN of the replication agreement.");
    fprintf(stderr, "\n\n-op controlreplerr -delete failure_ID -ra agreementDn");
    fprintf(stderr, "\n\n-op controlreplerr -retry  failure_ID -ra agreementDn");
    fprintf(stderr, "\n\nwhere:");
    fprintf(stderr, "\n     failure_ID  specifies the ID of the target replication update failure or '0' for all.");
    fprintf(stderr, "\n     agreementDn specifies the DN of the replication agreement.\n");

    exit(0);
}

void fullReplUsage() {
    fprintf(stderr, "\nThe usage for full replication operation is : ");
    fprintf(stderr, "\n\n-op fullRepl -ra agreementDn [options]");
    fprintf(stderr, "\nwhere:");
    fprintf(stderr, "\n agreementDn        DN of the agreement that identifies the consumer server, credentials, etc");
    fprintf(stderr, "\noptions:");
    fprintf(stderr, "\n -all               entire DIT will be replicated");
    fprintf(stderr, "\n -del               subtrees being replicated should be deleted from the consumer prior to replicating entries");
    fprintf(stderr, "\n -nofailIfExists    replication should continue if the consumer returns \'entry exists\'\n");
    exit(0);
}

void getUserTypeUsage() {
    fprintf(stderr, "\nThe usage for getusertype operation is : ");
    fprintf(stderr, "\n\n-op getusertype\n");
    exit(0);
}

void stopServerUsage() {
    fprintf(stderr, "\nThe usage for stopserver operation is : ");
    fprintf(stderr, "\n\n-op stopserver\n");
    exit(0);
}

/**********************************************************************
 * Function: acctStatusUsage
 * Purpose:  Print out a usage statement for the account status operation
 * Inputs: None
 * Pre-conditions: None
 * Post-conditions: program exits
 * Outputs: None          
 ***********************************************************************/
void acctStatusUsage() {
    fprintf(stderr, "\nThe usage for acctstatus operation is : ");
    fprintf(stderr, "\n\n-op acctstatus -d DN \n");
    exit(0);
}

void usage() {
    fprintf(stderr, "\nldapexop tool is used to perform extended operations.");
    fprintf(stderr, "\n\nusage:");
    fprintf(stderr, "\n ldapexop [general options] -op extendedOperation [operation specific options]");
    fprintf(stderr, "\n");
    fprintf(stderr, "\ngeneralOptions:");
    fprintf(stderr, "\n -C     charset   character set name to use, as registered with IANA");
    fprintf(stderr, "\n -d     level     set debug level to 'level' in LDAP library");
    fprintf(stderr, "\n -D     dn        bind dn");
    fprintf(stderr, "\n -e               display LDAP library version information and quit");
    fprintf(stderr, "\n -G     realm     realm used for DIGEST-MD5 bind mechanism");
    fprintf(stderr, "\n -h     host      LDAP server host name ");
    fprintf(stderr, "\n -K     keyfile   file to use for keys");
    fprintf(stderr, "\n -m     mechanism perform SASL bind with the given mechanism");
    fprintf(stderr, "\n -N     key_name  private key name to use in keyfile");
    fprintf(stderr, "\n -p     port      LDAP server port number");
    fprintf(stderr, "\n -P     key_pw    keyfile password");
    fprintf(stderr, "\n -U     username  user name for DIGEST-MD5 bind mechanism");
    fprintf(stderr, "\n -v               run in verbose mode");
    fprintf(stderr, "\n -w     password  bind password or '?' for non-echoed prompt");
    fprintf(stderr, "\n                  use backslash '\\?' to avoid matching single character filenames (UNIX only)");
    fprintf(stderr, "\n -x               use FIPS mode processing (SSL/TLS only)");
    fprintf(stderr, "\n -y     proxydn   set proxied id for proxied authorization operation");
    fprintf(stderr, "\n -Y               use a secure ldap connection (TLS)");
    fprintf(stderr, "\n -Z               use a secure ldap connection (SSL)");
    fprintf(stderr, "\n -?               display the usage");
    fprintf(stderr, "\n -help            display the usage");
    fprintf(stderr, "\n");
    fprintf(stderr, "\nextendedOperations: ");

    /*
     * Just required to uncomment the following lines to activate the help
     *  option of full replication
     */
    /* printf("\n fullRepl         start full replication extended operation");  */

    fprintf(stderr, "\n acctstatus       request the account status of given DN");
    fprintf(stderr, "\n cascrepl         cascading control replication extended operation");
    fprintf(stderr, "\n clearlog         clear log file extended operation");
    fprintf(stderr, "\n controlqueue     control queue extended operation");
    fprintf(stderr, "\n controlrepl      control replication extended operation");
    fprintf(stderr, "\n controlreplerr   control replication error log extended operation");
    fprintf(stderr, "\n evaluategroups   request the list of groups the user belongs to");
    fprintf(stderr, "\n getattributes    request attributes from atribute type extended operation");
    fprintf(stderr, "\n getlogsize       request log file size extended operation");
    fprintf(stderr, "\n getusertype      request user type extended operation");
    fprintf(stderr, "\n quiesce          quiesce or unquiesce subtree extended operation");
    fprintf(stderr, "\n readconfig       re-read configuration file extended operation");
    fprintf(stderr, "\n readlog          request lines from log file extended operation");
    fprintf(stderr, "\n repltopology     request to progagate replication topology entries");
    fprintf(stderr, "\n stopserver       request to stop the IBM Directory Server");
    fprintf(stderr, "\n unbind           unbind extended operation");
    fprintf(stderr, "\n uniqueattr       unique attribute extended operation\n");
    fprintf(stderr, "\nTo get help for a specific extended operation:");
    fprintf(stderr, "\nldapexop -op extendedOperation -help\n");
    exit(0);
}


/* This function checks whether the string entered is a number or not.*/
int is_num(char *s) {
    int rc = 0;
    if (s && (*s == '+' || *s == '-'))
        s++;
    while (s && *s) {
        if (!isdigit(*s))
            return rc;
        s++;
    }
    return 1;
}



/*  Get a password from the user but don't display it. */
char* getPassword( void ) {
    char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

    char in = '\0';                  /* Input character */
    int len = 0;                    /* Length of password */
#else

    struct termios echo_control;
    struct termios save_control;

    int fd = 0;                 /* File descriptor */
    int attrSet = 0;            /* Checked later for reset */

    /* Get the file descriptor associated with stdin. */
    fd = fileno( stdin );

    if (tcgetattr( fd, &echo_control ) != -1) {
        save_control = echo_control;
        echo_control.c_lflag &= ~( ECHO | ECHONL );

        if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
            fprintf(stderr, "Internal error setting terminal attribute.\n");
            exit( errno );
        }

        attrSet = 1;
    }
#endif

    /* Prompt for a password. */
    fputs( "Enter password ==> ", stdout );
    fflush( stdout );

#ifdef _WIN32
    /* Windows 9x/NT will always read from the console, i.e.,
       piped or redirected input will be ignored. */
    while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
        in = _getch();

        if (in != '\r') {
            supplied_password[len] = in;
            len++;

        } else {
            supplied_password[len] = '\0';
        }
    }
#else
    /* Get the password from stdin. */
    fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

    /* Remove the newline at the end. */
    supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
    /* Reset the terminal. */
    if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
        fprintf(stderr, "Unable to reset the display.\n");
    }
#endif

    fprintf( stdout, "\n" );

    return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}


/* This function will handle the portion of initializing a session with the server */
void initSession() {
    int failureReasonCode, authmethod;
    struct berval ber;
    struct berval *serverCreds = NULL;
    LDAPControl **pwdPolicy_controls = NULL;
    LDAPControl **bind_client_controls = NULL;

    if ((! (tls || ssl)) && keys) /* Specified a key file, but no security protocol */
        ssl = LDAP_TRUE;          /* Use SSL by default */

    /*Set the port*/
    if (port == 0) {
        if (ssl)
            port = LDAPS_PORT;
        else
            port = LDAP_PORT;
    }

    /* call ldap_ssl_client_init if SSL */
    if (ssl || tls) {

        /* Set FIPS mode processing if requested. */
        if ( use_FIPS_mode == LDAP_TRUE ) {
            if (verbose)
                printf( "ldap_ssl_set_fips_mode_np(1)\n");

            rc = ldap_ssl_set_fips_mode_np( 1 );
            if (rc != LDAP_SUCCESS) {
                fprintf( stderr,
                         "ldap_ssl_set_fips_mode_np failed\n" );
                exit( rc ) ;
            }
        }

        if ( !keyFile ) {
            keyFile = getenv("SSL_KEYRING");
            if (keyFile != NULL) {
                keyFile = strdup(keyFile);

                /* Make sure that we could allocate the memory for the keyFile variable. */
                if ( keyFile == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                } /*keyFile == NULL*/
            } /*keyFile != NULL*/
        } /*!keyFile*/

        if (verbose)
            printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                    ((keyFile) ? keyFile : "NULL"),
                    ((keyPwd) ? keyPwd : "NULL"));
        rc = ldap_ssl_client_init( keyFile, keyPwd, 0, &failureReasonCode );
        if (rc != LDAP_SUCCESS) {
            fprintf( stderr,
                     "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                     rc, failureReasonCode );
            fprintf( stderr, "%s\n", ssl_err2string( failureReasonCode ));
            exit( 1 );
        } /*rc != LDAP_SUCCESS*/
        if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n", ((host) ? host : "NULL"), port,
                   ((keyName) ? keyName : "NULL"));

        /* If we're using SSL, do an ldap_ssl_init. Otherwise (TLS), use the normal ldap_init */
        if (ssl)
            ld = ldap_ssl_init(host, port, keyName );
        else
            ld = ldap_init(host, port );

        if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed\n" );
            perror( "ldap_ssl_init" );
            exit( 1 );
        }
    } else { /*!ssl*/
        if (verbose)
            printf("ldap_init(%s, %d) \n", ((host) ? host : "NULL"), port);
        if ((ld = ldap_init(host, port)) == NULL) {
            perror("ldap_init");
            exit(1);
        }
    } /*ssl*/

    /* set options */
    ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
    ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&debugLevel);
    /* don't chase referrals during the bind */
    ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )LDAP_OPT_OFF);

    if (charset != NULL) {
        if ( verbose )
            printf( "ldap_set_iconv_local_charset( %s )\n", charset );

        if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
            fprintf(stderr, "Unsupported charset %s\n", charset);
            exit(0);
        }

        ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
    } /*charset != NULL*/

    if (tls) {                   /* Start the TLS session */
        if ((rc = ldap_start_tls_s_np(ld, NULL)) != LDAP_SUCCESS) {
            fprintf(stderr, "ldap_start_tls_s() failed! RC = %d\n", rc);
            fprintf(stderr, "ldap_start_tls_s_np: %s\n", ldap_err2string(ldap_get_errno(ld)));
            exit(rc);
        }
    } /*tls*/

    if ( !mechanism ) /* Use simple bind */
    {
        if (verbose)
            printf( "ldap_simple_bind_s( ld, %s, %s )\n",
                    (dn ? dn : "NULL"),
                    (passwd ? passwd : "NULL"));

        rc = ldap_simple_bind_s(ld, dn, passwd);
        if ( rc != LDAP_SUCCESS ) {
            ldap_perror( ld, "ldap_simple_bind" );
            /*
            ** LDAP_OPT_EXT_ERROR only valuable for ssl communication.
            ** In this example, for LDAP v3, the bind is the first
            ** instance in which communication actually flows to the
            ** server.  So, if there is an ssl configuration error or
            ** other ssl problem, this will be the first instance where
            ** it will be detected.
            */
            if (ssl) {
                ldap_get_option( ld, LDAP_OPT_EXT_ERROR, &failureReasonCode);
                fprintf( stderr, "Attempted communication over SSL.\n");
                fprintf( stderr, "  The extended error is %d.\n", failureReasonCode);
            }
            exit( rc );
        } /*rc != LDAP_SUCCESS*/
    } /*!mechanism*/
    else /* Presence of mechanism means SASL bind */ {
        /*
        ** Special case for mechanism="EXTERNAL".  Unconditionally set bind DN
        ** and credentials to NULL.  This option should be used in tandem
        ** with SSL and client authentication.  For other SASL mechanisms,
        ** use the specified bind DN and credentials.
        */
        if (strcmp(mechanism, LDAP_MECHANISM_EXTERNAL) == 0) {

            if (verbose)
                printf( "ldap_sasl_bind_s( ld, NULL, %s, NULL ,NULL,NULL,serverCreds)\n",
                        (mechanism ? mechanism : "NULL"));

            rc = ldap_sasl_bind_s (ld, NULL, mechanism, NULL, NULL, NULL, &serverCreds);
            if (rc != LDAP_SUCCESS ) {
                ldap_perror ( ld, "ldap_sasl_bind_s" );
                exit ( rc );
            }

        } else {
            /*   Set password policy control*/
            if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &pwdPolicy_controls)) != LDAP_SUCCESS) {
                fprintf( stderr, "ldap_add_control: %d, %s\n",
                         rc, ldap_err2string( rc ) );
                exit(rc);
            }
            if (strcmp(mechanism, LDAP_MECHANISM_GSSAPI) == 0) {
                if (verbose)
                    printf( "ldap_sasl_bind_s( ld, NULL, %s, NULL ,NULL,NULL,serverCreds)\n",
                            (mechanism ? mechanism : "NULL"));

                rc = ldap_sasl_bind_s (ld, NULL, mechanism, NULL, pwdPolicy_controls, NULL, &serverCreds);  /* Use pwdPolicy controls */
                if (rc != LDAP_SUCCESS ) {
                    ldap_perror ( ld, "ldap_sasl_bind_s" );
                    exit ( rc );
                }

            } else /* other SASL mechanisms */ {
                if (strcmp(mechanism, LDAP_MECHANISM_DIGEST_MD5) == 0) {
                    if (digest_username) {
                        if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_USER_NAME_OID, strlen(digest_username) , digest_username, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                            ldap_perror (ld, "ldap_add_control");
                            exit ( rc );
                        }
                    }
                    if (digest_realm) {
                        if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_REALM_NAME_OID, strlen(digest_realm) , digest_realm, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                            ldap_perror (ld, "ldap_add_control");
                            exit ( rc );
                        }
                    }
                }

                ber.bv_len = strlen ( passwd );
                ber.bv_val = passwd;
                if (verbose)
                    printf( "ldap_sasl_bind_s( ld, NULL, %s, NULL ,NULL,NULL,serverCreds)\n",
                            (mechanism ? mechanism : "NULL"));

                rc = ldap_sasl_bind_s (ld, dn, mechanism, &ber, pwdPolicy_controls, bind_client_controls, &serverCreds); /* use UTF8 and pwdPolicy controls */
                if (rc != LDAP_SUCCESS ) {
                    ldap_perror ( ld, "ldap_sasl_bind_s" );
                    exit ( rc );
                }
                /*rc != LDAP_SUCCESS*/

                /* during the operation, chase referrals */
                ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )LDAP_OPT_ON);

                /*Free the password policy controls. */
                ldap_controls_free(pwdPolicy_controls);

                /*Free the client controls*/
                ldap_controls_free(bind_client_controls);
            } /* other SASL mechanisms */
        } /*else*/
    } /*SASL*/
}

/* ssl_err2string returns a string describing the error associated with
** the given error code from an SSL related call.
*/
const char * ssl_err2string( int ssl_error ) {
    switch ( ssl_error ) {
    case LDAP_SSL_EXTENDED_OK:
        return LDAP_SSL_EXTENDED_OK_STR;
    case LDAP_SSL_EXTENDED_INVALID_HANDLE:
        return LDAP_SSL_EXTENDED_INVALID_HANDLE_STR;
    case LDAP_SSL_EXTENDED_API_NOT_AVAILABLE:
        return LDAP_SSL_EXTENDED_API_NOT_AVAILABLE_STR;
    case LDAP_SSL_EXTENDED_INTERNAL_ERROR:
        return LDAP_SSL_EXTENDED_INTERNAL_ERROR_STR;
    case LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE:
        return LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE_STR;
    case LDAP_SSL_EXTENDED_INVALID_STATE:
        return LDAP_SSL_EXTENDED_INVALID_STATE_STR;
    case LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND:
        return LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND_STR;
    case LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE:
        return LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE_STR;
    case LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION:
        return LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION_STR;
    case LDAP_SSL_EXTENDED_ERROR_CRYPTO:
        return LDAP_SSL_EXTENDED_ERROR_CRYPTO_STR;
    case LDAP_SSL_EXTENDED_ERROR_ASN:
        return LDAP_SSL_EXTENDED_ERROR_ASN_STR;
    case LDAP_SSL_EXTENDED_ERROR_LDAP:
        return LDAP_SSL_EXTENDED_ERROR_LDAP_STR;
    case LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR:
        return LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR_STR;
    case LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR:
        return LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR_STR;
    case LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR:
        return LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR_STR;
    case LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT:
        return LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT_STR;
    case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY:
        return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY_STR;
    case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL:
        return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL_STR;
    case LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD:
        return LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD_STR;
    case LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED:
        return LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED_STR;
    case LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB:
        return LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB_STR;
    case LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR:
        return LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR_STR;
    case LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD:
        return LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD_STR;
    case LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR:
        return LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR_STR;
    case LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR:
        return LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR_STR;
    case LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT:
        return LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT_STR;
    case LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS:
        return LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_INDEX:
        return LDAP_SSL_EXTENDED_ERROR_BAD_INDEX_STR;
    case LDAP_SSL_EXTENDED_CLOSE_FAILED:
        return LDAP_SSL_EXTENDED_CLOSE_FAILED_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_DATE:
        return LDAP_SSL_EXTENDED_ERROR_BAD_DATE_STR;
    case LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS:
        return LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS_STR;
    case LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE:
        return LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE:
        return LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE_STR;
    case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE:
        return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE_STR;
    case LDAP_SSL_EXTENDED_ERROR_IO:
        return LDAP_SSL_EXTENDED_ERROR_IO_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL:
        return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD:
        return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT:
        return LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE:
        return LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_MAC:
        return LDAP_SSL_EXTENDED_ERROR_BAD_MAC_STR;
    case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED:
        return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG:
        return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_CERT:
        return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_PEER:
        return LDAP_SSL_EXTENDED_ERROR_BAD_PEER_STR;
    case LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED:
        return LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED_STR;
    case LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED:
        return LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED_STR;
    case LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION:
        return LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION_STR;
    case LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION:
        return LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION_STR;
    case LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED:
        return LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER:
        return LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER:
        return LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE:
        return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_STR;
    case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION:
        return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION_STR;
    case LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED:
        return LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED_STR;
    case LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED:
        return LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED_STR;
    case LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE:
        return LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE_STR;
    case LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY:
        return LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY_STR;
    case LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED:
        return LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED_STR;
    case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH:
        return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH_STR;
    case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT:
        return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT_STR;
    case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD:
        return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD_STR;
    case LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER:
        return LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER_STR;
    case LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE:
        return LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE_STR;
    case LDAP_SSL_EXTENDED_WOULD_BLOCK:
        return LDAP_SSL_EXTENDED_WOULD_BLOCK_STR;
    case LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3:
        return LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3_STR;
    case LDAP_SSL_EXTENDED_MISC_INVALID_ID:
        return LDAP_SSL_EXTENDED_MISC_INVALID_ID_STR;
    case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID:
        return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID_STR;
    case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH:
        return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH_STR;
    case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION:
        return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION_STR;
    case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE:
        return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE_STR;
    case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE:
        return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE_STR;
    case LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING:
        return LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING_STR;
    default:
        return "Unknown SSL error";
    }
}


/*Functions to create the Request Value for the extended operation*/

/* This function is used to ecode the input data for a read log request */
/*
 * linesFrom specifies the starting position of the log file.
 * linesTo specifies the ending position of the log file.
 */
struct berval *createRequestReadlog(int logNumber) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{iii}\",%d,%d,%d)\n", logNumber, linesFrom, linesTo);

    /* Compose the ber element */
    if (ber_printf(ber, "{iii}", logNumber, linesFrom, linesTo) == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/* This function is used to encode the input data for a log request (get number of lines and clear)*/
/*
 *  logNumber specifies the log number to be used for encoding :
 * 1 - Slapd
 * 2 - Cli
 * 4 - Audit
 * 8 - Bulkload
 * 16 - ibmdiradm
 * 32 - adminAudit
 * 64 - debug
 * 128 - LostAndFound
 * 256 - Configuration Tools log
 */
struct berval *createRequestLog(int logNumber) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{i}\",%d)\n", logNumber);

    if (ber_printf(ber, "{i}", logNumber) == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/* This function is used to encode the input data for a readconfig request.*/
/*
 * scopeNum specifies the scope to be used for encoding :
 *
 * 0 - Entire
 * 1 - Specific Entry
 *
 *
 *
 * entryDn   : is basically the DN of the entry under which the modification
 *    is to be done.
 * attribute : is the actual attribute getting re-read.
 */
struct berval *createRequestReadConfig(int scopeNum) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{iss}\",%d,%s,%s)\n", scopeNum, entryDn, attribute);

    /* Compose the ber element here */
    if (ber_printf(ber, "{iss}", scopeNum, entryDn ? entryDn : "", attribute ? attribute : "") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/* This function is used to encode the input data for a cascrepl request*/
/*
 * actionNum : The action to be performed as shown below:
 *     0. Quiesce
 *     1. Unquiesce
 *     2. Replicate Now
 *     3. Wait for replication.
 * contextDn : The DN of the subtree-root to be affected.
 * timeout  : Maximum allowable time for this operation to complete.
 */
struct berval *createRequestCascRepl(int actionNum) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{isi}\",%d,%s,%d)\n", actionNum, contextDn, timeout);

    /* Compose the ber element here */
    if (ber_printf(ber, "{isi}", actionNum, contextDn ? contextDn : "", timeout) == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/* This function is used to encode the input data for a controlqueue request*/
/*
 * actionNum : The action to be performed as shown below:
 *     0. Skip All.
 *     1. Skip Single.
 * agreementDn : The DN of the agreement to be affected.
 * changeId : The Change Id to the skipped.
 */
struct berval *createRequestControlQ(int actionNum) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{iss}\",%d,%s,%s)\n", actionNum, agreementDn, changeId);

    /* Compose the ber element here */
    if (ber_printf(ber, "{iss}", actionNum, agreementDn ? agreementDn : "", changeId ? changeId : "") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/* This function is used to encode the input data for a controlrepl request*/
/*
 * actionValue : The action to be performed as shown below:
 *     0. Suspend.
 *     1. Resume.
 *     2. ReplicateNow.
 *     3. TerminateFullReplication.
 *
 * scopeValue : The scope of the action to be performed:
 *     0. Single agreement.
 *     1. All agreements associated with the replication agreement.
 *
 * contextDn  : The DN of the agreement/subtree root to be considered.
 */
struct berval *createRequestControlRepl(int actionValue, int scopeValue) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{iis}\",%d,%d,%s)\n", actionValue, scopeValue, contextDn);

    /* Compose the ber element here */
    if (ber_printf(ber, "{iis}", actionValue, scopeValue, contextDn ? contextDn : "") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/* This function is used to encode the input data for a quiesce request*/
/*
 *  contextDn  : The DN of the subtree root to be [un]quiesced
 *  end    : false -> unquiesce.
 *      true -> quiesce.
 */
struct berval *createRequestQuiesce() {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{is}\",%d,%s)\n", !end, contextDn);

    /* Compose the ber element here */
    if (ber_printf(ber, "{is}", !end, contextDn ? contextDn : "") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}

/* This function is used to encode the input data for a repltopology request*/
/*
 * contextDn : The DN of the subtree-root to be affected.
 * timeout  : Maximum allowable time for this operation to complete.
 * agreementDn : The DN of the agreement to be affected.
 */
struct berval *createRequestReplTopology() {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{sis}\",%s,%d,%s)\n", contextDn, timeout, agreementDn);

    /* Compose the ber element here */
    if (ber_printf(ber, "{sis}", contextDn ? contextDn : "", timeout, agreementDn ? agreementDn: "") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}

/* This function is used to encode the input data for a fullrepl request*/
/*
 * contextDn  : The DN of the subtree root to be [un]quiesced
 * end    : false -> unquiesce.
 *      true -> quiesce.
 */
struct berval *createRequestFullRepl() {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{siii}\",%s,%d,%d,%d)\n", agreementDn, all, del, nofailIfExists);

    /* Compose the ber element here */
    if (ber_printf(ber, "{siii}", agreementDn ? agreementDn : "", all, del, nofailIfExists) == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}

/* This function is used to encode the input data for a getusertype request*/
struct berval * createRequestGetUserType() {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}

/* This function is used to encode the input data for stopserver extended operation.
 * The encoded data will have an "option" representing STOP and and blank set of parameters
 * The method returns a berval struct.
 *
 */
struct berval *createRequestStopServer(int option, const char* clOpts) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to a berval. */

    if (verbose)
        printf("ber_alloc(1)\n");

    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{is}\",%i,%s)\n", option, "");

    if (ber_printf( ber, "{is", option, clOpts) == (LBER_ERROR)) {
        fprintf( stderr, "first ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    if (ber_printf(ber, "}") == (LBER_ERROR)) {
        fprintf( stderr, "closing ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/*
 * This function is used to encode the input data for an unbind request.
 *
 */

struct berval *createRequestUnbind(int uDNValue, int uIPValue) {
    struct berval *ret = NULL; /* Used for storing the encoded form to be returned to the main routine. */
    BerElement *ber = NULL;    /* Used for storing the encoded form to be converted to the binary form i.e.
                                                                      * struct berval format*/

    if (verbose)
        printf("ber_alloc(1)\n");

    /* Allocate the memory for one sequence */
    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    /* Compose the ber element here */
    if (ber_printf(ber, "{") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /*if only a uDNSelected*/
    if (uDNValue == 1 && uIPValue != 2) {
        if (verbose)
            printf("ber_printf(ber, \"{{es}}\",%d,%s)\n", uDNValue, udn);

        if (ber_printf(ber, "{es}", uDNValue, udn) == (LBER_ERROR)) {
            fprintf( stderr, "ber_printf failed\n");
            ber_free(ber, 1);
            return NULL;
        }
    }
    /*if only a uIPSelected*/
    else if (uIPValue == 2 && uDNValue != 1) {

        if (verbose)
            printf("ber_printf(ber, \"{{es}}\",%d,%s)\n", uIPValue, uip);

        if (ber_printf(ber, "{es}", uIPValue, uip) == (LBER_ERROR)) {
            fprintf( stderr, "ber_printf failed\n");
            ber_free(ber, 1);
            return NULL;
        }
    }
    /* If both DN and IP are NOT selected. -all case*/
    else if (uDNValue != 1 && uIPValue != 2) {
        if (verbose)
            printf("ber_printf(ber, \"{}\",)\n");

    }
    /* If both DN and IP are selected*/
    else if (uDNValue == 1 && uIPValue == 2) {
        if (verbose)
            printf("ber_printf(ber, \"{{es}{es}}\",%d,%s,%d,%s)\n", uDNValue, udn, uIPValue, uip);

        if (ber_printf(ber, "{es}{es}", uDNValue, udn, uIPValue, uip) == (LBER_ERROR)) {
            fprintf( stderr, "ber_printf failed\n");
            ber_free(ber, 1);
            return NULL;
        }
    }

    /* always need to have a close } */
    if (ber_printf(ber, "}") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    /* Encode the ber element into BER format */
    if (ber_flatten(ber, &ret) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/* This function is used to encode the input data into a berval structure
 *
 * @ return : berval struct
 */

struct berval * createRequestUniqueAttr() {
    struct berval *ret = NULL;
    BerElement *ber = NULL;

    if (verbose)
        printf("ber_alloc(1)\n");

    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{s}\",%s)\n", attrType);

    if (ber_printf( ber, "{s}", attrType) == ( -1)) {
        fprintf( stderr, "first ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    if (ber_flatten(ber, &ret) == ( -1)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}

/* This function is used to encode the input data into a berval structure
 *
 */

struct berval * createRequestGetAttrs(int attrNum) {
    struct berval *ret = NULL;
    BerElement *ber = NULL;

    if (verbose)
        printf("ber_alloc(1)\n");

    if ((ber = ber_alloc_t(1)) == NULL) {
        fprintf( stderr, "ber_alloc_t failed\n");
        return NULL;
    }

    if (verbose)
        printf("ber_printf(ber, \"{ib}\",%d,%d)\n", attrNum, iMatch);

    if (ber_printf( ber, "{ib}", attrNum, iMatch) == ( -1)) {
        fprintf( stderr, "first ber_printf failed\n");
        ber_free(ber, 1);
        return NULL;
    }

    if (ber_flatten(ber, &ret) == ( -1)) {
        fprintf( stderr, "ber_flatten failed\n");
        ber_free(ber, 1);
        return NULL;
    }
    ber_free(ber, 1);
    return ret;
}


/*****************/
/*-op getusertype*/
/*****************/

/* This function is used for decoding a berval structure, having an o/p. as an array of strings and int. */
/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 *  return  : int 0 for Success and -1 for failure
 */
int decodeBerGetUserType(struct berval * result,
                         char ** userType,                                  /*Will contain only one value*/
                         int * numUserRoles,                                   /* Number of UserTypes returned from the server */
                         char *** userRoles) /* Sequence of UserRoles returned from the server */ {
    int response = 0;  /* Stores the return code of the BER decoding function */

    if (result) {
        BerElement *ber = ber_init2(result);  /* Obtain the BerElement corresponding to the BER encoded format. */
        response = ber_scanf(ber, "{aiv", userType, numUserRoles, userRoles);
        if (verbose)
            printf("ber_scanf(ber,\"{aiv\",userType, numUserRoles,userTypes)\n");

        if (response == -1) {
            fprintf(stderr, "\nBer_scanf failed");
            return -1;
        } else
            return 0;
    } else {
        fprintf(stderr, "Response from Server Missing.\n");
        return -1;
    }
}

/* This function performs the getuserype extended operation.
 *
 * The user types returned can be anyone or more of the following:
 * "root_administrator",
 *  "admin_group_member",
 *  "ldap_user"
 */

void getUserType() {
    struct berval *reqVal; /*This will be the BER encoded form of the data sent to the server*/
    const char *oid = LDAP_USER_TYPE_REQ_OID;    /*The oid for the request */
    char *oidres = NULL;   /*The oid received in response from the server*/
    struct berval *valres = NULL; /*The BER encoded response from the server*/

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    char * userType = NULL; /*String containing the userType*/
    char ** userRoles = NULL;  /*Sequence of strings containing the User Roles*/
    char ** prevUserRoles = NULL;
    int numUserRoles = 0;   /* Number of Roles */
    int count = 0;
    int decode_rc = 0;

    /* Start a session with the server */
    initSession();

    reqVal = createRequestGetUserType();

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decode_rc = decodeBerGetUserType(valres, &userType, &numUserRoles, &userRoles); /* Get the User Types*/

        printf("\nUser    : %s\n", userType);
        if (decode_rc != 0) {
            fprintf( stderr, "\n Error in getting the user roles.\n");
            exit(rc);
        } else if (numUserRoles < 1) {
            fprintf( stderr, "\n No roles were returned from the Server.\n");
        } else /* User roles are returned from the server */ {
            /* print the User roles*/
            printf("Role(s) : ");
            for (count = 0;count < numUserRoles;count++) {
                printf("%s ", *userRoles);
                prevUserRoles = userRoles++;

                /* free the String  */
                ldap_memfree(*prevUserRoles);
            }
            printf("\n");
        }
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

        rc = ldap_get_lderrno(ld, &dn, &msg);
        if ( msg != NULL && msg[0] != 0 )
            printf ( "msg: %s\n", msg);
    }

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}

/*****************/
/*-op stopserver*/
/*****************/

/* This function performs the stopserver extended operation.
 *
 */
void stopServer() {
    struct berval *reqVal; /*This will be the BER encoded form of the data sent to the server*/
    const char *oid = LDAP_ADMIN_SLAPD;    /*The oid for the request */
    char *oidres = NULL;   /*The oid received in response from the server*/
    struct berval *valres = NULL; /*The BER encoded response from the server*/

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestStopServer(STOP, ""); /*We don't send any options on a stop request*/

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        printf("Stop operation succeeded\n" );
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else 
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);

}
/*Given a logName this functions returns a corresponding log type number*/
int getLogNumber(char *logName) {
    /* Obtain the corresponding log number */
    if (!strcmp(logName, "slapd"))
        return SlapdErrors;
    else if (!strcmp(logName, "cli"))
        return CLIErrors;
    else if (!strcmp(logName, "audit"))
        return AuditLog;
    else if (!strcmp(logName, "bulkload"))
        return BulkloadLog;
    else if (!strcmp(logName, "ibmdiradm"))
        return AdminErrors;
    else if (!strcmp(logName, "debug"))
        return DebugOutputFile;
    else if (!strcmp(logName, "LostAndFound"))
        return LostAndFoundLog;
    else if (!strcmp(logName, "config"))
        return ConfigToolsLog;
    else
        return AdminAudit;
}

/*****************/
/*-op clearlog   */
/*****************/

/* This function clears the log
 */
void clearLog() {
    struct berval *reqVal; /* This will be the BER encoded form of the data sent to the server */
    const char *oid = LDAP_LOG_CLEAR_REQ_OID;   /* The oid for clearing the specified log file */
    char *oidres = NULL;   /* The oid received in response fromt the server */
    struct berval *valres = NULL; /* The BER encoded response from the server */
    int logNumber;    /* This will store the lognumber referring to a log name */

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* Start a session with the server */
    initSession();

    logNumber = getLogNumber(logName);

    /* Create the request */
    reqVal = createRequestLog(logNumber);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        printf("\n%s log file cleared.\n", logName);
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}

/**
 *    This function handles all the input parameters for the "clearlog" operation
 **/

int parseClearLogInputs(int argIndex, int argc, char *argv[], Option *list) {
    int logRequiredFlag = 0;  /* This flag is used to check if the "log" option is specified or not */


    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) /* In case the current option is a valid option */
        {
            switch ( optVal->optionID ) {
            case CLEARLOGLOG:                                                 /* -log  */
                logRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    clearlogUsage();
                logName = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the logName variable. */
                if ( logName == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;

                /* Check for the value of "-log"..it should be either audit,bulkload,cli or slapd */

                if ( strcmp(logName, "audit") != 0
                        && strcmp(logName, "bulkload") != 0
                        && strcmp(logName, "cli") != 0
                        && strcmp(logName, "slapd") != 0
                        && strcmp(logName, "ibmdiradm") != 0
                        && strcmp(logName, "debug") != 0
                        && strcmp(logName, "adminAudit") != 0
                        && strcmp(logName, "LostAndFound") != 0
                        && strcmp(logName, "config") != 0) {
                    clearlogUsage();
                }


                /* The variables to be used are :
                   log
                */

                clearLog();   /* Call the function to clear the log */

                break;

            case OPTHELP:                                      /* help  */
                clearlogUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                clearlogUsage();
                rc = -1;
                break;

            default:
                break;
            }

        }
        else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            clearlogUsage();
            rc = -1;
        }


    }

    /* Check if the mandatory arguments are specified by the user or not */

    if (logRequiredFlag == 0 ) {
        clearlogUsage();
    }
    return argIndex;

}

/* This function is used for decoding a berval structure, having an o/p. as an integer. */
/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 */

int decodeBerGetLogsize(struct berval * result) {
    int response;  /* Stores the return code of the BER decoding function */
    int numLines;  /* To store the number of lines returned from the server */
    if (result) {
        BerElement *ber = ber_init2(result);  /* Obtain the BerElement corresponding to the BER encoded format. */
        response = ber_scanf(ber, "i", &numLines); /* Scan the "number of lines" returned by the server */
        if (verbose)
            printf("ber_scanf(ber,\"i\",&numLines)\n");

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            return -1;
        } else
            return numLines;
    } else {
        fprintf(stderr, "Server not responding.\n");
        return -1;
    }
}


/* This function is used for decoding a berval structure, having an o/p. as an octet string. */

/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 */

void decodeBerReadlog(struct berval * result) {
    int response;  /* Stores the return code of the BER decoding function */
    if (result) {
        BerElement *ber = ber_init2(result); /* Obtain the BerElement corresponding to the BER encoded format. */
        char * resultString = NULL;   /* To store the string(s) returned from the server */

        response = ber_scanf(ber, "a", &resultString);

        if (verbose)
            printf("ber_scanf(ber,\"a\",&resultString)\n");

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            return;
        }
        printf("%s", resultString);
    } else {
        fprintf(stderr, "Server not responding.\n");
    }
}

/*
 * This function adjusts the variabls linesFrom and linesTo w.r.t to the
 * max. size of the file so that the requested no. of lines from the server
 * can be reduced.
 * e.g.
 * If file size is 10 lines and the input parameters are to read the
 * file from 1 to 1000, then the upper limit is reduced to 10.
 */

void manipulateLines(int logNumber) {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_LOG_NUMLINES_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;
    int fileSize; /* Stores the number of lines in the current log file */

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* The assumption here is that the session is already initialized,
     * i.e. ld != NULL
     */

    /* Create the request */
    reqVal = createRequestLog(logNumber);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        fileSize = decodeBerGetLogsize(valres);
        if (fileSize < 0) {
            fprintf(stderr, "\n%s log not present.\n", logName);
            exit(rc);
        }

        if (linesFrom > fileSize - 1) /* In case the file size is less than the lower limit specified */
        {
            printf("\nNo lines to display in this range.\n");
            exit(0);
        } else
            if (linesTo > fileSize - 1)  /* In case the file size is more than the upper limit specified */
                linesTo = fileSize - 1;
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

        ldap_get_lderrno(ld, &dn, &msg);
        if ( msg != NULL && msg[0] != 0 )
            printf ( "msg: %s\n", msg);

        exit(rc);
    }

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);

}

/*****************/
/*-op readlog    */
/*****************/

/* This function reads the log specified through the log number from
 * line number linesFrom to line number linesTo
 */

void readLog() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_LOG_GETLINES_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    /* This will store the lognumber as compared to the log name. */
    int logNumber;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    logNumber = getLogNumber(logName);

    /* Start a session with the server */
    initSession();

    if (strcmp(lines, "all") == 0) {
        linesFrom = 0; /* Initialize the linesFrom = 0 and linesTo = no. of lines in the file */

        /* Create the request */
        reqVal = createRequestLog(logNumber);

        oid = LDAP_LOG_NUMLINES_REQ_OID;

        /* Issue the operation */
        rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

        oid = LDAP_LOG_GETLINES_REQ_OID;

        /* Check our result */
        if (rc == LDAP_SUCCESS) {
            linesTo = decodeBerGetLogsize(valres); /* Get the log size */
            ldap_memfree(oidres);
            ber_bvfree(valres);
            if (linesTo < 0) {
                fprintf( stderr, "\n%s log not present.\n", logName);
                exit(rc);
            }
        } else {
            fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

            ldap_get_lderrno(ld, &dn, &msg);
            if ( msg != NULL && msg[0] != 0 )
                printf ( "msg: %s\n", msg);

            exit(rc);
        }
    } else {
        manipulateLines(logNumber); /* Adjust the line-limits appropriately */
    }

    if (reqVal) {
        ber_bvfree(reqVal);
        reqVal = NULL;
    }

    /* Create the request */
    reqVal = createRequestReadlog(logNumber);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerReadlog(valres);
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}


/**
 *    This function handles all the input parameters for the "readlog" operation
 **/

int parseReadLogInputs(int argIndex, int argc, char *argv[], Option *list) {
    int logRequiredFlag = 0;    /* This flag is used to check if the "log" option is specified or not */
    int linesRequiredFlag = 0;  /* This flag is used to check if the "lines" option is specified or not */
    int numDigits;      /* To store the number of digits after the preceding zeroes(if any). */
    /* i.e. for 0011 numDigits = 2 */

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {

            case READLOGLOG:                                                 /* -log */

                logRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex]) {
                    readlogUsage();
                }
                logName = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the logName variable. */
                if ( logName == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;

                /* Check for the value of "-log"..it should be either audit,bulkload,cli or slapd */

                if ( strcmp(logName, "audit") != 0
                        && strcmp(logName, "bulkload") != 0
                        && strcmp(logName, "cli") != 0
                        && strcmp(logName, "slapd") != 0
                        && strcmp(logName, "ibmdiradm") != 0
                        && strcmp(logName, "debug") != 0
                        && strcmp(logName, "adminAudit") != 0
                        && strcmp(logName, "LostAndFound") != 0
                        && strcmp(logName, "config") != 0) {
                    readlogUsage();
                }

                break;

            case READLOGLINES:                                                 /* -lines */

                linesRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    readlogUsage();
                lines = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the lines variable. */
                if ( lines == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;

                /* Check for the value */

                if ( strcmp(lines, "all") != 0 ) {
                    Inputlength = strlen(lines);

                    /* now check if all chars are integers in the input string..... */

                    for (i = 0; i < Inputlength; i++) {
                        if (isdigit(lines[i]) == 0) {
                            /* To deal with the case '+1' and '+' */
                            /* i.e. -lines +1 10 is allowed but not + 10 */
                            if (i == 0 && Inputlength > 1 && lines[i] == '+')
                                continue;
                            readlogUsage();
                        }
                    }

                    /* Obtain the number of digits excluding the preceding 0s */
                    if (lines[0] == '+')
                        numDigits = Inputlength - strspn(&lines[1], "0") - 1;
                    else
                        numDigits = Inputlength - strspn(lines, "0");

                    linesFrom = atoi(lines);

                    /* In case the number of digits > 10, flash an error message */

                    if ( linesFrom < 0 || numDigits > 10) {
                        fprintf(stderr, "\n\"linesFrom\" parameter exceeds limit.");
                        fprintf(stderr, "\nMaximum value of 2147483647 expected.\n");
                        /* This limit has been decided by means of a test in the form of numbers > 0x7fffffff
                         * This indicates that the server expects this parameter in the form of a signed int.
                         * Hence the necessary care is taken at the client side itself.
                         */
                        exit(0);
                    }

                    if (lines) {
                        free (lines);   /* Since we are going to overwrite lines here
                        * better free it before using it further. */
                    }

                    /* Get the other integer for "linesTo" */

                    if (!argv[argIndex])
                        readlogUsage();
                    lines = strdup(argv[argIndex]);

                    /* Make sure that we could allocate the memory for the lines variable. */
                    if ( lines == NULL) {
                        fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                        exit(1);
                    }

                    argIndex++;

                    Inputlength = strlen(lines);


                    /* now check if all chars are integers in the input string..... */

                    for (i = 0; i < Inputlength; i++) {
                        if (isdigit(lines[i]) == 0) {
                            /* To deal with the case '+1' and '+' */
                            /* i.e. -lines +1 10 is allowed but not + 10 */
                            if (i == 0 && Inputlength > 1 && lines[i] == '+')
                                continue;
                            readlogUsage();
                        }
                    }

                    /* Obtain the number of digits excluding the preceding 0s */
                    if (lines[0] == '+')
                        numDigits = Inputlength - strspn(&lines[1], "0") - 1;
                    else
                        numDigits = Inputlength - strspn(lines, "0");

                    linesTo = atoi(lines);

                    if ( linesTo < 0 || numDigits > 10) {
                        fprintf(stderr, "\n\"linesTo\" parameter exceeds limit.");
                        fprintf(stderr, "\nMaximum value of 2147483647 expected.\n");
                        /* This limit has been decided by means of a test in the form of numbers > 0x7fffffff
                         * This indicates that the server expects this parameter in the form of a signed int.
                         * Hence the necessary care is taken at the client side itself.
                         */
                        exit(0);
                    }

                    if (linesTo < linesFrom) {
                        fprintf(stderr, "\n\"linesTo\" should be greater then or equal to \"linesFrom\".\n");
                        exit(0);
                    }
                }
                break;
            case OPTHELP:                                      /* help  */
                readlogUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                readlogUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex]);
            readlogUsage();
            rc = -1;
        }
    }

    /* Check if the mandatory arguments are specified by the user or not */

    if (linesRequiredFlag == 0 || logRequiredFlag == 0) {
        readlogUsage();
    }

    readLog(); /* Call the routine for reading the log file */
    return argIndex;
}


/*****************/
/*-op getlogsize*/
/*****************/

/* This function gives the size of the log specified by logNumber
*/
void getLogSize() {
    struct berval *reqVal;
    const char *oid = LDAP_LOG_NUMLINES_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;
    int numLines; /* To store the number of lines in the log file */
    int logNumber; /* To store the log number corresponding to the specified log file*/

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* Start a session with the server */
    initSession();

    logNumber = getLogNumber(logName);

    /* Create the request */
    reqVal = createRequestLog(logNumber);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        numLines = decodeBerGetLogsize(valres);
        if (numLines >= 0) {
            printf("\n%ld line%c", numLines, numLines != 1 ? 's' : '\0');
            printf("\n");
        } else
            fprintf(stderr, "\n%s log not present.\n", logName);
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}


/**
 *    This function handles all the input parameters for the "getlogsize" operation
 **/
int parseLogSizeInputs(int argIndex, int argc, char *argv[], Option *list) {
    int logRequiredFlag = 0;  /* This flag is used to check if the "log" option is specified or not */

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {

            case LOGSIZELOG:                                                 /* -log */

                logRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    getlogsizeUsage();
                logName = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the logName variable. */
                if ( logName == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;

                /* Check for the value of "-log"..it should be either audit,bulkload,cli or slapd */

                if ( strcmp(logName, "audit") != 0
                        && strcmp(logName, "bulkload") != 0
                        && strcmp(logName, "cli") != 0
                        && strcmp(logName, "slapd") != 0
                        && strcmp(logName, "ibmdiradm") != 0
                        && strcmp(logName, "debug") != 0
                        && strcmp(logName, "adminAudit") != 0
                        && strcmp(logName, "LostAndFound") != 0
                        && strcmp(logName, "config") != 0) {
                    getlogsizeUsage();
                }

                break;

            case OPTHELP:                                      /* help  */
                getlogsizeUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                getlogsizeUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            getlogsizeUsage();
            rc = -1;
        }
    }

    /* Check if the mandatory arguments are specified by the user or not */

    if (logRequiredFlag == 0 ) {
        getlogsizeUsage();
    }

    getLogSize();
    return argIndex;
}

/*****************/
/*-op readconfig*/
/*****************/

/*
 * This function re-reads the configuration file as per the following values :
 * scope = entire file or a specific attribute or a specfic entry or a subtree.
 * entryDn = the DN which will have the attribute or which needs to be re-read.
 * attribute = the actual attribute to be modified.
 */
void readConfig() {
    struct berval *reqVal;
    const char *oid = LDAP_CONFIG_UPDATE_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /*Integer representation of the scope value*/
    int scopeNum;

    /* Start a session with the server */
    initSession();

    scopeNum = iScope;    /* assign the scope */

    /* Create the request */
    reqVal = createRequestReadConfig(scopeNum);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        printf("\nConfiguration file successfully updated.\n");
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);

}

/**
 *    This function handles all the input parameters for the "readconfig" operation
 **/

int parseReadConfigInputs(int argIndex, int argc, char *argv[], Option *list) {

    int scopeRequiredFlag = 0;  /* This flag is used to check if the "scope" option is specified or not */

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {


            case READCONFIGSCOPE:                                      /* -scope */

                scopeRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    readconfigUsage();
                scope = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the scope variable. */
                if ( scope == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                argIndex++;

                /* Check for the value of "-scope"..it should be anyone of these: "single","entire" or "subtree" */
                if ( strcmp(scope, SINGLE_SCOPE) != 0
                        && strcmp(scope, ALL_SCOPE) != 0
                        && strcmp(scope, ENTRY_SCOPE) != 0
                        && strcmp(scope, SUBTREE_SCOPE) != 0) {
                    readconfigUsage();
                }

                /* Now get the "entryDn" and the "attribute"....*/
                if (strcmp(scope, SINGLE_SCOPE) == 0) {
                    /* Check for the entry Dn*/
                    if (!argv[argIndex])
                        readconfigUsage();
                    entryDn = strdup(argv[argIndex++]);

                    /* Make sure that we could allocate the memory for the entryDn variable. */
                    if ( entryDn == NULL) {
                        fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                        exit(1);
                    }

                    /* Check for the attribute*/
                    if (argv[argIndex]) {
                        attribute = strdup(argv[argIndex++]);

                        /*set the scope*/
                        iScope = ATTRIBUTE;

                        /* Make sure that we could allocate the memory for the attribute variable. */
                        if (attribute == NULL) {
                            fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                            exit(1);
                        }
                    } else {
                        readconfigUsage();
                    }
                }


                /* Now get the "entry"*/
                if (strcmp(scope, ENTRY_SCOPE) == 0) {

                    /* Check for the entry Dn*/
                    if (!argv[argIndex])
                        readconfigUsage();
                    entryDn = strdup(argv[argIndex++]);

                    /* Make sure that we could allocate the memory for the entryDn variable. */
                    if ( entryDn == NULL) {
                        fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                        exit(1);
                    }

                    /*set the scope*/
                    iScope = ENTRY;
                }

                /* Now get the "subtree"*/
                if (strcmp(scope, SUBTREE_SCOPE) == 0) {

                    /*set the scope*/
                    iScope = SUBTREE;

                    /* Check for the entry Dn*/
                    if (!argv[argIndex])
                        readconfigUsage();
                    entryDn = strdup(argv[argIndex++]);

                    /* Make sure that we could allocate the memory for the entryDn variable. */
                    if ( entryDn == NULL) {
                        fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                        exit(1);
                    }

                    /* Check for any further arguments*/
                    if (argv[argIndex]) {
                        readconfigUsage();
                    }
                }

                break;
            case OPTHELP:                                      /* help  */
                readconfigUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                readconfigUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            readconfigUsage();
            rc = -1;
        }
    }

    /* Check if the mandatory arguments are specified by the user or not */

    if (scopeRequiredFlag == 0 ) {
        readconfigUsage();
    }

    /* The attributes to be used are :
     *   scope
     *   entryDn
     *   attribute
     */

    readConfig();

    return argIndex;
}

/*
 * Helper function for unexpected errorCode from replication ex ops
 */
void printOtherError(int resultCode, const char *errorMessage) {
    fprintf(stderr, "Operation failed with result code %d.", resultCode);
    if (errorMessage && *errorMessage)
        fprintf(stderr, " Details: %s", errorMessage);
    fprintf(stderr, "\n");
}

/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 */

void decodeBerCascadeRepl(const char * oidres, struct berval * result) {
    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_REPL_CASC_CONTROL_REQ_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {
        BerElement *ber = ber_init2(result); /* Obtain the BerElement corresponding to the BER encoded format. */
        int resultCode;
        char * msg;
        int response;
        char * supplier = NULL;
        char * consumer = NULL;

        response = ber_scanf(ber, "{iaaa", &resultCode, &msg, &supplier, &consumer);

        if (verbose) {
            printf("ber_scanf(ber,\"{iaa\",&resultCode,&msg,&supplier,&consumer)\n");
            if (response != -1) {
                printf("  resultCode=%d\n", resultCode);
                printf("  msg=%s\n", msg);
                printf("  supplier=%s\n", supplier);
                printf("  consumer=%s\n", consumer);
            }
        }

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }
        /* print consumer/supplier info if there was an error */
        if (resultCode != LDAP_SUCCESS) {
            if (*consumer) {
                printf("Server %s reported an error from server %s.\n",
                       supplier, consumer);
            } else {
                fprintf(stderr, "Operation failed on server %s.\n",
                        supplier);
            }
        }
        switch (resultCode) {
        case LDAP_SUCCESS:
            printf("\nOperation completed successfully.\n");
            break;
        case LDAP_INSUFFICIENT_ACCESS:
            fprintf(stderr, "You do not have authority to perform this operation.\n");
            break;
        case LDAP_TIMEOUT:
            fprintf(stderr, "Operation did not complete within specified time.\n");
            break;
        default:
            printOtherError(resultCode, msg);
            break;
        }
        ldap_memfree(msg);
        ldap_memfree(supplier);
        ldap_memfree(consumer);
        ber_free(ber, 1);
    }
}

/*****************/
/*-op cascrepl   */
/*****************/

/*
 * This function initiates cascaded replication as per the following values :
 *
 * action:  Quiesce/Unquiesce/replicateNow/waitForReplication.
 * subtreeDN: Identifies the subtree which this operation affects.
 * timeout: Specifies the maximum number of seconds to wait for the
 *    operation to complete.
 */

void cascadeRepl() {
    struct berval *reqVal;
    const char *oid = LDAP_REPL_CASC_CONTROL_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    /*Action to perform*/
    int actionNum = 3;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* Start a session with the server */
    initSession();

    if (strcmp(action, "quiesce") == 0)
        actionNum = LDAP_REPL_CASCADE_QUIESCE; /* Quiesce. */
    else if (strcmp(action, "unquiesce") == 0)
        actionNum = LDAP_REPL_CASCADE_UNQUIESCE; /* Unquiesce. */
    else if (strcmp(action, "replnow") == 0)
        actionNum = LDAP_REPL_CASCADE_REPLICATE_NOW; /* Replicate now. */

    /* Create the request */
    reqVal = createRequestCascRepl(actionNum);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS)
        decodeBerCascadeRepl(oidres, valres);
    else
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);

}


/**
 *    This function handles all the input parameters for the "cascrepl" operation
 **/

int parseCascadeInputs(int argIndex, int argc, char *argv[], Option *list) {
    int actionRequiredFlag = 0;  /* This flag is used to check if the "action" option is specified or not */
    int rcRequiredFlag = 0;      /* This flag is used to check if the "rc" option is specified or not */
    char *timeoutS = NULL;  /* This variable stores the string value corresponding to timeout */
    int numDigits;    /* Stores the actual no. of digits excluding the preceding 0s and the probable + sign */


    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {

            case CASCREPLACTION:                                                 /* -action */

                actionRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    cascreplUsage();
                action = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the action variable. */
                if ( action == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;

                /* Check for the value of "-action"..it should be either quiesce,unquiesce,wait or replnow*/
                if ( strcmp(action, "quiesce") != 0 && strcmp(action, "unquiesce") != 0
                        && strcmp(action, "replnow") != 0 && strcmp(action, "wait") != 0 ) {
                    cascreplUsage();
                }


                /*       printf("\nAction = %s",action); */

                break;

            case CASCREPLTIMEOUT:                                      /* -timeout */

                argIndex++;
                if (!argv[argIndex])
                    cascreplUsage();

                timeoutS = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the timeoutS variable. */
                if ( timeoutS == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                Inputlength = strlen(timeoutS);

                /* now check if all chars are integers in the input string..... */

                for (i = 0; i < Inputlength; i++) {
                    if (isdigit(timeoutS[i]) == 0) {
                        /* To deal with the case '+1' and '+' */
                        /* i.e. -lines +1 10 is allowed but not + 10 */
                        if (i == 0 && Inputlength > 1 && timeoutS[i] == '+')
                            continue;
                        cascreplUsage();
                    }
                }

                /* Obtain the number of digits excluding the preceding 0s */

                if (timeoutS[0] == '+')
                    numDigits = Inputlength - strspn(&timeoutS[1], "0") - 1;
                else
                    numDigits = Inputlength - strspn(timeoutS, "0");

                timeout = atoi(timeoutS);
                free(timeoutS);
                /* In case the number of digits > 10, flash an error message */

                if ( timeout < 0 || numDigits > 10) {
                    fprintf(stderr, "\n\"timeout\" parameter exceeds limit.");
                    fprintf(stderr, "\nMaximum value of 2147483647 expected.\n");
                    /* This limit has been decided by means of a test in the form of numbers > 0x7fffffff
                     * This indicates that the server expects this parameter in the form of a signed int.
                     * Hence the necessary care is taken at the client side itself.
                     */
                    exit(0);
                }

                argIndex++;
                break;

            case CASCREPLRC:                                          /* -rc */
                rcRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    cascreplUsage();
                contextDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the contextDn variable. */
                if ( contextDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case OPTHELP:                                      /* help  */
                cascreplUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                cascreplUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported Extended Operation option: %s.\n", argv[argIndex] );
            cascreplUsage();
            rc = -1;
        }


    }

    /* Check if the mandatory arguments are specified by the user or not */

    if (rcRequiredFlag == 0 || actionRequiredFlag == 0 ) {

        cascreplUsage();
    }

    /* Attributes required are :
     * action
     * contextDn
     * timeout
     */

    cascadeRepl();
    return argIndex;

}

/******************/
/*-op controlqueue*/
/******************/

/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 */

void decodeBerControlQ( struct berval * result) {
    int response;  /* Stores the return code of the BER decoding function */
    if (result) {
        BerElement *ber = ber_init2(result); /* Obtain the BerElement corresponding to the BER encoded format. */
        int resultCode;
        char * msg = NULL;
        int skippedCount;

        response = ber_scanf(ber, "{iai", &resultCode, &msg, &skippedCount);

        if (verbose) {
            printf("ber_scanf(ber,\"{iai\",&resultCode,&msg,&skippedCount)\n");
            if (response != -1) {
                printf("  resultCode=%d\n", resultCode);
                printf("   msg=%s\n", msg);
                printf("   skippedCount=%d\n", skippedCount);
            }
        }

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }
        switch (resultCode) {
        case LDAP_SUCCESS:
            if (skippedCount == 0)
                printf("Operation completed, but there were no pending changes.\n");
            else
                printf("%d changes skipped.\n", skippedCount);
            break;
        case LDAP_INSUFFICIENT_ACCESS:
            fprintf(stderr, "You do not have authority to perform this operation.\n");
            break;
        case LDAP_UNWILLING_TO_PERFORM:
            fprintf(stderr, "Specified change is not the next change to replicate."
                    " The server may already have replicated this change.\n");
            break;
        default:
            printOtherError(resultCode, msg);
            break;
        }
        ldap_memfree(msg);
        ber_free(ber, 1);
    } else {
        fprintf(stderr, "Server not responding.\n");
    }
}

/*
 * This function controls the replication queue as per the following values :
 *
 * skip  : 'all' to skip all changes.
 * agreementDN : Identifies the agreementDn which this operation affects.
 */

void controlQueue() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_REPL_CONTROL_QUEUE_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    int actionValue = LDAP_REPL_QUEUE_SKIPSINGLE;  /* To store the number corresponding to the action to be performed */

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    if (strcmp(skip, "all") == 0)
        actionValue = LDAP_REPL_QUEUE_SKIPALL;
    else {
        changeId = skip;
    }

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestControlQ(actionValue);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerControlQ(valres);
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal != NULL)
        ber_bvfree(reqVal);
    if (oidres != NULL)
        ldap_memfree(oidres);
    if (valres != NULL)
        ber_bvfree(valres);
}



/**
 *    This function handles all the input parameters for the "controlqueue" operation
 **/

int parseControlQueueInputs(int argIndex, int argc, char *argv[], Option *list) {
    int raRequiredFlag = 0;   /* This flag is used to check if the "-ra" option is specified or not */
    int skipRequiredFlag = 0; /* This flag is used to check if the "-skip" option is specified or not */

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {
            case CONTROLQUEUESKIP:                                                 /* -skip */

                skipRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    controlqueueUsage();

                skip = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the skip variable. */
                if ( skip == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case CONTROLQUEUERA:                                         /* -ra */

                raRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    controlqueueUsage();
                agreementDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the agreementDn variable. */
                if ( agreementDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;
            case OPTHELP:                                      /* help  */
                controlqueueUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                controlqueueUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            controlqueueUsage();
            rc = -1;
        }


    }

    if ( raRequiredFlag == 0 || skipRequiredFlag == 0) {
        controlqueueUsage();
    }

    /* The values which will be used are :
     * skip
     * agreementDn
     */

    controlQueue();

    return argIndex;

}

/*****************/
/*-op controlrepl*/
/*****************/

/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 */

void decodeBerControlReplication(const char * oidres, struct berval * result) {
    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_REPL_CONTROL_REPL_REQ_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {
        BerElement *ber = ber_init2(result); /* Obtain the BerElement corresponding to the BER encoded format. */
        int resultCode;
        char * msg = NULL;
        int response;
        char * consumer = NULL;

        response = ber_scanf(ber, "{iaa", &resultCode, &msg, &consumer);

        if (verbose) {
            printf("ber_scanf(ber,\"{iaa\",&resultCode,&msg,&consumer)\n");
            if (response != -1) {
                printf("  resultCode=%d\n", resultCode);
                printf("  msg=%s\n", msg);
                printf("  consumer=%s\n", consumer);
            }
        }

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }
        if ((resultCode != LDAP_SUCCESS) && consumer && *consumer)
            printf("Error occurred with agreement to server %s.\n", consumer);
        switch (resultCode) {
        case LDAP_SUCCESS:
            printf("\nOperation completed successfully.\n");
            break;
        case LDAP_INSUFFICIENT_ACCESS:
            fprintf(stderr, "You do not have authority to perform this operation.\n");
            break;
        default:
            printOtherError(resultCode, msg);
            break;
        }
        ldap_memfree(msg);
        ldap_memfree(consumer);
        ber_free(ber, 1);
    }
}

/*
 * This function controls the replication process as per the following values :
 *
 * action  : 0 -> suspend
 *     1 -> resume
 *     2 -> replicateNow
 *     3 -> terminateFullReplication
 * scope  : single -> single agreement
 *     all  ->  all the existing agreements pertaining to the subtree specified by entryDN.
 * contextDn : entry to be affected.
 */

void controlReplication() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_REPL_CONTROL_REPL_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    int actionValue;    /* To store the number corresponding to the action to be performed */
    int scopeValue = LDAP_REPL_CONTROL_REPL_SCOPE_SINGLE_AGREEMENT; /* To store the number corresponding to the scope specified. */

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    if (strcmp(action, "suspend") == 0)
        actionValue = LDAP_REPL_CONTROL_REPL_ACTION_SUSPEND;
    else if (strcmp(action, "resume") == 0)
        actionValue = LDAP_REPL_CONTROL_REPL_ACTION_RESUME;
    else if (strcmp(action, "replnow") == 0)
        actionValue = LDAP_REPL_CONTROL_REPL_ACTION_REPLICATENOW;
    else
        actionValue = 3;

    if (strcmp(scope, "all") == 0)
        scopeValue = LDAP_REPL_CONTROL_REPL_SCOPE_ALL_AGREEMENTS;

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestControlRepl(actionValue, scopeValue);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerControlReplication(oidres, valres);
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind*/
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}

/**
 *    This function handles all the input parameters for the "controlrepl" operation
 **/

int parseControlReplInputs(int argIndex, int argc, char *argv[], Option *list) {
    int actionRequiredFlag = 0; /* This flag is used to check if the "action" option is specified or not*/
    /*   int scopeRequiredFlag = 0; */ /* This flag is used to check if the "scope" option is specified or not */
    int raRequiredFlag = 0;     /* This flag is used to check if the "ra" option is specified or not */
    int rcRequiredFlag = 0;     /* This flag is used to check if the "ra" option is specified or not */


    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {

            case CONTROLREPLACTION:                                                 /* -action */

                actionRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    controlreplUsage();
                action = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the action variable. */
                if ( action == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;

                /* Check for the value of "-action"..it should be either suspend,resume or replflow */
                /* To be uncommented when full replication is to be supported.
                 * if ( strcmp(action,"suspend")!=0 && strcmp(action,"resume")!=0
                 *  && strcmp(action,"replnow")!=0 && strcmp(action,"terminate")!=0 )
                 * {
                 *   controlreplUsage();
                 * }
                 */
                if ( strcmp(action, "suspend") != 0 && strcmp(action, "resume") != 0
                        && strcmp(action, "replnow") != 0) {
                    controlreplUsage();
                }

                break;

            case CONTROLREPLRA:                                          /* -ra */
                raRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    controlreplUsage();
                contextDn = strdup(argv[argIndex]);
                scope = strdup("single");

                /* Make sure that we could allocate the memory for the scope variable. */
                if ( scope == NULL || contextDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;

                break;

            case CONTROLREPLRC:
                rcRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    controlreplUsage();
                contextDn = strdup(argv[argIndex]);
                scope = strdup("all");

                /* Make sure that we could allocate the memory for the scope variable. */
                if ( scope == NULL || contextDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case OPTHELP:                                      /* help  */
                controlreplUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                controlreplUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            controlreplUsage();
            rc = -1;
        }


    }

    /* Check if the mandatory arguments are specified by the user or not */

    if ( (raRequiredFlag == rcRequiredFlag) /* both or neither specified -- exactly one must be specified */
            || (actionRequiredFlag == 0)) {
        controlreplUsage();
    }

    /* Required variables */
    /* action
     * scope
     * contextDN
     */

    controlReplication();

    return argIndex;

}

/******************/
/*-op repltopology*/
/******************/
/*
 * result : Stores the BER encoded value returned from the server as a result of repltopology extended
 *    operation.
 */

void decodeBerReplTopology(const char * oidres, struct berval * result) {
    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_REPL_TOPOLOGY_RES_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {
        BerElement *ber = ber_init2(result); /* Obtain the BerElement corresponding to the BER encoded format. */
        int resultCode;
        char * msg;
        int response;
        char * supplier = NULL;
        char * consumer = NULL;

        response = ber_scanf(ber, "{iaaa", &resultCode, &msg, &supplier, &consumer);

        if (verbose) {
            printf("ber_scanf(ber,\"{iaa\",&resultCode,&msg,&supplier,&consumer)\n");
            if (response != -1) {
                printf("  resultCode=%d\n", resultCode);
                printf("  msg=%s\n", msg);
                printf("  supplier=%s\n", supplier);
                printf("  consumer=%s\n", consumer);
            }
        }
        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }
        /* print consumer/supplier info if there was an error */
        if (resultCode != LDAP_SUCCESS) {
            if (*consumer) {
                printf("Server %s reported an error from server %s.\n",
                       supplier, consumer);
            } else {
                fprintf(stderr,"Operation failed on server %s.\n",
                        supplier);
            }
        }
        switch (resultCode) {
        case LDAP_SUCCESS:
            if (msg && *msg != '\0') {
                printf("\nOperation completed successfully. Details: %s\n", msg);
            } else {
                printf("\nOperation completed successfully.\n");
            }
            break;
        case LDAP_INSUFFICIENT_ACCESS:
            fprintf(stderr,"You do not have authority to perform this operation.\n");
            break;
        case LDAP_TIMEOUT:
            if (msg && *msg != '\0') {
                fprintf(stderr,"Operation did not complete within specified time. Details: %s\n", msg);
            } else {
                fprintf(stderr,"Operation did not complete within specified time.\n");
            }
            break;
        default:
            printOtherError(resultCode, msg);
            break;
        }
        ldap_memfree(msg);
        ldap_memfree(supplier);
        ldap_memfree(consumer);
        ber_free(ber, 1);
    }
}

/*
 * This function initiates topology replication as per the following values :
 *
 * subtreeDN: Identifies the subtree which this operation affects.
 * timeout: Specifies the maximum number of seconds to wait for the
 *          operation to complete.
 * agreementDN : Identifies the agreementDn which this operation affects.
 *          This is optional.
 */
void replTopology() {
    struct berval *reqVal;
    const char *oid = LDAP_REPL_TOPOLOGY_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestReplTopology();

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, NULL, NULL, &oidres, &valres);
    /* Check our result */
    if (rc == LDAP_SUCCESS)
        decodeBerReplTopology(oidres, valres);
    else
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);

}

/**
 *    This function handles all the input parameters for the "repltopology" operation
 **/
int parseReplTopologyInputs(int argIndex, int argc, char *argv[], Option *list) {
    int raRequiredFlag = 0;  /* This flag is used to check if the "ra" option is specified or not */
    int rcRequiredFlag = 0;  /* This flag is used to check if the "rc" option is specified or not */
    char *timeoutS = NULL;   /* This variable stores the string value corresponding to timeout */
    int numDigits;    /* Stores the actual no. of digits excluding the preceding 0s and the probable + sign */


    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {

            case REPLTOPOTIMEOUT:                                    /* -timeout */
                argIndex++;
                if (!argv[argIndex])
                    replTopologyUsage();

                timeoutS = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the timeoutS variable. */
                if ( timeoutS == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                Inputlength = strlen(timeoutS);

                /* now check if all chars are integers in the input string..... */

                for (i = 0; i < Inputlength; i++) {
                    if (isdigit(timeoutS[i]) == 0) {
                        /* To deal with the case '+1' and '+' */
                        /* i.e. -lines +1 10 is allowed but not + 10 */
                        if (i == 0 && Inputlength > 1 && timeoutS[i] == '+')
                            continue;
                        replTopologyUsage();
                    }
                }
                /* Obtain the number of digits excluding the preceding 0s */

                if (timeoutS[0] == '+')
                    numDigits = Inputlength - strspn(&timeoutS[1], "0") - 1;
                else
                    numDigits = Inputlength - strspn(timeoutS, "0");

                timeout = atoi(timeoutS);
                free(timeoutS);
                /* In case the number of digits > 10, flash an error message */

                if ( timeout < 0 || numDigits > 10) {
                    fprintf(stderr, "\n\"timeout\" parameter exceeds limit.");
                    fprintf(stderr, "\nMaximum value of 2147483647 expected.\n");
                    /* This limit has been decided by means of a test in the form of numbers > 0x7fffffff
                     * This indicates that the server expects this parameter in the form of a signed int.
                     * Hence the necessary care is taken at the client side itself.
                     */
                    exit(0);
                }

                argIndex++;
                break;

            case REPLTOPOLOGYRC:                                        /* -rc */
                rcRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    replTopologyUsage();
                contextDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the contextDn variable. */
                if ( contextDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;
            case REPLTOPOLOGYRA:                                        /* -ra */
                raRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    replTopologyUsage();
                agreementDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the agreementDn variable. */
                if ( agreementDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case OPTHELP:                                    /* help  */
                replTopologyUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                   /* ? */
                replTopologyUsage();
                rc = -1;
                break;

            default:
                break;
            }
        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported Extended Operation option: %s.\n", argv[argIndex] );
            replTopologyUsage();
            rc = -1;
        }


    }
    /* Check if the mandatory arguments are specified by the user or not */

    if (rcRequiredFlag == 0 ) {

        replTopologyUsage();
    }
    /* Attributes required are :
     * contextDn
     * timeout
     */

    replTopology();
    return argIndex;

}

/* decodeBerUnbind */
/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 */

void
decodeBerUnbind(const char * oidres, struct berval * result) {
    int numOfConKilled = 0;
    int numOfConPending = 0;

    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_DN_UNBIND_RES_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {
        BerElement *ber = ber_init2(result); /* Obtain the BerElement corresponding to the BER encoded format. */
        int resultCode;
        char * msg;
        int response;



        response = ber_scanf(ber, "{ii}", &numOfConKilled, &numOfConPending);

        if (verbose) {
            printf("ber_scanf(ber,\"{ii}\",&numOfConKilled, &numofConPending)\n");

            if (response != -1) {
                printf ("\nNumber of connections killed  : %d", numOfConKilled);
                printf ("\nNumber of connections pending : %d", numOfConPending);
            }

        }

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            return;
        } else {

            printf ("\nNumber of connections killed  : %d", numOfConKilled);
            printf ("\nNumber of connections pending : %d\n", numOfConPending);
        }

        ber_free(ber, 1);
    }
}


/*
 * Function to unbind connections
 *
 */

void
unbind(int uDNSel, int uIPSel) {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_DN_UNBIND_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    /*used by ldap_get_lderrno*/
    char *dn = NULL;
    char *msg = NULL;

    int iKilled = 0;
    int iPending = 0;

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestUnbind(uDNSel, uIPSel);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerUnbind(oidres, valres);
    } else {
        fprintf( stderr, "Operation Did Not Complete Successfully.  RC=%i\n", rc);
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);

}



/**
 *
 *    This function handles all the input parameters for the "unbind" operation
 *
 **/

int
parseUnbindInputs(int argIndex, int argc, char *argv[], Option *list) {
    int uDNSelected = 0; /* This flag is used to check if the "dn" option is specified or not*/
    int uIPSelected = 0;     /* This flag is used to check if the "ip" option is specified or not */
    int noneSelected = 1; /* This flag is used to check if no option is selcted.*/

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {

            case UNBINDDN:                                                /* -dn */
                noneSelected = 0;
                argIndex++;
                if (!argv[argIndex])
                    unbindUsage();
                udn = strdup(argv[argIndex]);


                /* Make sure that we could allocate the memory for the dn variable. */
                if ( udn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                uDNSelected = 1;
                argIndex++;
                break;
            case UNBINDIP:                                  /* -ip */

                noneSelected = 0;
                argIndex++;
                if (!argv[argIndex])
                    unbindUsage();
                uip = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the scope variable. */
                if ( uip == NULL ) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                uIPSelected = 2;
                argIndex++;
                break;

            case UNBINDALL:

                noneSelected = 0;
                uIPSelected = 0;
                uDNSelected = 0;
                argIndex++;
                udn = strdup("");
                uip = strdup("");

                /* Make sure that we could allocate the memory for the scope variable. */
                if ( udn == NULL || uip == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                break;

            case OPTHELP:                                  /* help  */
                unbindUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                  /* ? */
                unbindUsage();
                rc = -1;
                break;
            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            unbindUsage();
            rc = -1;
        }


    }

    /* Check if the mandatory arguments are specified by the user or not */

    if (noneSelected) {
        unbindUsage();
    }

    /* Required variables */
    /* dn
     * ip
     * all
     */

    unbind(uDNSelected, uIPSelected);

    return argIndex;

}

/*****************/
/*-op quiesce    */
/*****************/

/*
 * result : Stores the BER encoded value returned from the server as a result of an extended
 *    operation.
 */

void decodeBerQuiesce(const char * oidres, struct berval * result) {
    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_REPL_QUIESCE_REQ_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {
        BerElement *ber = ber_init2(result); /* Obtain the BerElement corresponding to the BER encoded format. */
        int resultCode;
        char * msg;
        int response;

        response = ber_scanf(ber, "{ia", &resultCode, &msg);

        if (verbose) {
            printf("ber_scanf(ber,\"{ia\",&resultCode,&msg)\n");
            if (response != -1) {
                printf("  resultCode=%d\n", resultCode);
                printf("  msg=%s\n", msg);
            }
        }

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            return;
        }
        switch (resultCode) {
        case LDAP_SUCCESS:
            printf("\nOperation completed successfully.\n");
            break;
        case LDAP_INSUFFICIENT_ACCESS:
            fprintf(stderr, "You do not have authority to perform this operation.\n");
            break;
        case LDAP_REPL_QUIESCE_BAD_STATE:
            fprintf(stderr, "Replication context is already in the requested state.\n");
            break;
        default:
            printOtherError(resultCode, msg);
            break;
        }
        ldap_memfree(msg);
        ber_free(ber, 1);
    }
}

/*
 * Function to quiesce a server over a specific DN
 * Variables to be used:
 * contextDn -> The DN of the subtree root to be [un]quiesced.
 * end   -> false  -> quiesce the subtree
 *     true  -> unquiesce the subtree
 */
void quiesce() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_REPL_QUIESCE_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestQuiesce();

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerQuiesce(oidres, valres);
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}


/**
 *    This function handles all the input parameters for the "quiesce" operation
 **/

int parseQuiesceInputs(int argIndex, int argc, char *argv[], Option *list) {
    int rcRequiredFlag = 0;  /* This flag is used to check if the "-rc" option is specified or not */

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {


            switch ( optVal->optionID ) {
            case QUIESCERC:                                                 /* -rc */

                rcRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    quiesceUsage();
                contextDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the contextDn variable. */
                if ( contextDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case QUIESCEEND:                                      /* -end */
                end = LDAP_TRUE;
                argIndex++;
                break;
            case OPTHELP:                                      /* help  */
                quiesceUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                quiesceUsage();
                rc = -1;
                break;

            default:
                break;
            }
        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            quiesceUsage();
            rc = -1;
        }
    }

    if ( rcRequiredFlag == 0 ) {
        quiesceUsage();
    }

    /* Required attributes :
     * contextDn
     * end
     */

    quiesce();

    return argIndex;
}

/* This function need not be considered for the current release.
 * This has been implemented with future releases into consideration.
 */


/*
 * Function to do the full replication extended operation
 *
 * Variables to be used:
 *
 *  agreementDn  -> The DN of the agreement to be taken up for full replication.
 *  all    -> true -> Take up all the agreements for full replication
 *       false -> Take up the agreement specified above.
 *  del    -> true -> Delete the subtree on the target server before replication.
 *       false -> Do not delete any entry pre-existing on the target server, prior to replication.
 *  -> true -> Do not fail in case the entry to be replicated exists on the target server.
 *       false -> Fail in case the entry to be replicated already exists on the target server.
 */
void fullRepl() {
    struct berval *reqVal = NULL;
    const char *oid = "1.3.18.0.2.12.18";
    char *oidres = NULL;
    struct berval *valres = NULL;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    /* Make sure that we could allocate the memory for the oid variable. */
    if ( oid == NULL) {
        fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
        exit(1);
    }

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestFullRepl();

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        printf("\nOperation completed successfully.\n");
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 )
        printf ( "msg: %s\n", msg);

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}



/**
 *    This function handles all the input parameters for the "fullRepl" operation
 **/

int parseFullReplInputs(int argIndex, int argc, char *argv[], Option *list) {
    int requiredFlag = 0;  /* This flag is used to check if the "-ra" option is specified or not */
    int delFlag = 0;       /* This flag is used to check is the "-del" option is specified or not. */
    int noFailFlag = 0;    /* This flag is used to check is the "-nofailIfExists" option is specified or not.*/

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {
            case FULLREPAGGDN:                                              /* -ra */

                requiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    fullReplUsage();
                agreementDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the agreementDn variable. */
                if ( agreementDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case FULLREPALL:                                      /* -all */
                all = LDAP_TRUE;
                argIndex++;
                break;

            case FULLREPDEL:                                      /* -del */
                delFlag = 1;
                del = LDAP_TRUE;
                argIndex++;
                break;

            case FULLREPNOFAIL:                                     /* -nofailIfExists */
                noFailFlag = 1;
                nofailIfExists = LDAP_TRUE;
                argIndex++;
                break;
            case OPTHELP:                                      /* help  */
                fullReplUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                fullReplUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            fullReplUsage();
            rc = -1;
        }
    }

    if (requiredFlag == 0) {
        fullReplUsage();
    }


    /* Required variables

    agreementDn
    all
    delete
    nofailIfExists

    */

    fullRepl();

    return argIndex;

}
/*****************/
/*-op uniqueattr */
/*****************/

/* This function is used for decoding a berval structure for uniqueattr extended operation,
 * having an o/p. as an array of strings.
 *
 */

void decodeBerUniqueAttr(const char * oidres,
                         struct berval * result,
                         char *** attrValues) /* Sequence of attribute values returned from the server */ {
    int response;  /* Stores the return code of the BER decoding function */
    /*   int resultCode;
       char * msg;  */

    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_UNIQUE_ATTR_RES_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {

        BerElement *ber = ber_init2(result);  /* Obtain the BerElement corresponding to the BER encoded format. */
        response = ber_scanf(ber, "{v}", attrValues);
        if (verbose)
            printf("ber_scanf(ber,\"{v}\", attrValues)\n");

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }

        /*      if (resultCode != LDAP_SUCCESS) {
                 switch (resultCode) {
                 case LDAP_INSUFFICIENT_ACCESS:
                    printf("You do not have authority to perform this operation.\n");
                    break;
                 default:
                    printOtherError(resultCode, msg);
                    break;
                 }
              } */
    }
}

/*
 * Function to get all conflicting values for a specific attributeType
 */

void uniqueAttr() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_UNIQUE_ATTR_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;
    int index = 0;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    char ** attrValues = NULL;

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = createRequestUniqueAttr();

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerUniqueAttr(oidres, valres, &attrValues);

        if (attrValues == NULL) {
            printf ( "\nNo conflicting attribute values returned.\n");

        } else {
            printf("Attribute values are :\n");
            for (index = 0; attrValues != NULL && attrValues[index] != NULL; index++) {
                printf ( "%s \n", attrValues[index]);
                ldap_memfree(attrValues[index]);
            }
        }
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

        rc = ldap_get_lderrno(ld, &dn, &msg);
        if ( msg != NULL && msg[0] != 0 )
            printf ( "msg: %s\n", msg);
    }

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}


/**
 *    This function handles all the input parameters for the "uniqueattr" operation
**/

int parseUniqueAttrInputs(int argIndex, int argc, char *argv[], Option *list) {

    int aRequiredFlag = 0; /* This flag is used to check if "-a" option is specified or not */

    while ( (argIndex < argc) && (rc == 0) ) {

        if (isOption( argv[argIndex], list) ) {
            switch ( optVal->optionID ) {
            case ATTRTYPE:                                                 /* -a */
                aRequiredFlag = 1;
                argIndex++;
                if (!argv[argIndex])
                    uniqueAttrUsage();
                attrType = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the attrType variable. */
                if ( attrType == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case OPTHELP:                                      /* help  */
                uniqueAttrUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                uniqueAttrUsage();
                rc = -1;
                break;

            default:
                break;
            }
        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            uniqueAttrUsage();
            rc = -1;
        }
    }

    if ( aRequiredFlag == 0 ) {
        uniqueAttrUsage();
    }

    uniqueAttr();

    return argIndex;
}

/********************/
/*-op getattributes */
/********************/

/* This function is used for decoding a berval structure for getattributes extended operation,
 * having an o/p. as an array of strings.
 *
 */

void decodeBerGetAttrs(const char * oidres,
                       struct berval * result,
                       char *** attrValues) /* Sequence of attribute values returned from the server */ {
    int response;  /* Stores the return code of the BER decoding function */
    /*   int resultCode;
       char * msg;  */

    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_ATTR_TYPE_RES_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {

        BerElement *ber = ber_init2(result);  /* Obtain the BerElement corresponding to the BER encoded format. */
        response = ber_scanf(ber, "{v}", attrValues);
        if (verbose)
            printf("ber_scanf(ber,\"{v}\", attrValues)\n");

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }
    }
}

/*Given an attrtype this functions returns a corresponding enumeration number*/
int getAttributeNumber(char *attributeType) {
    /* Obtain the corresponding attribute type number */
    if (!strcmp(attributeType, "operational"))
        return OPERATIONAL;
    else
        if (!strcmp(attributeType, "language_tag"))
            return LANGUAGE_TAG;
        else
            if (!strcmp(attributeType, "attribute_cache"))
                return ATTRIBUTE_CACHE;
            else
                if (!strcmp(attributeType, "unique"))
                    return UNIQUE_ATTR;
                else
                    return CONFIGURATION;
}

/*
 * Function to implement getAttributes extended operation
 */

void getAttributes() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_ATTR_TYPE_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;
    int index = 0;
    int attrNumber = 0;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    char ** attrValues = NULL;

    /* Start a session with the server */
    initSession();

    /*get the attributeType*/
    attrNumber = getAttributeNumber(attrsType);

    /* Create the request */
    reqVal = createRequestGetAttrs(attrNumber);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerGetAttrs(oidres, valres, &attrValues);

        if (attrValues == NULL) {
            printf ( "\nNo values returned from the Server.\n");

        } else {
            printf("Attribute values are :\n");
            for (index = 0; attrValues != NULL && attrValues[index] != NULL; index++) {
                printf ( "%s \n", attrValues[index]);
                ldap_memfree(attrValues[index]);
            }
        }
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

        rc = ldap_get_lderrno(ld, &dn, &msg);
        if ( msg != NULL && msg[0] != 0 )
            printf ( "msg: %s\n", msg);
    }

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}

/**
 *    This function handles all the input parameters for the "getattributes" operation
**/

int parseGetAttrsInputs(int argIndex, int argc, char *argv[], Option *list) {

    int iAttrType = 0; /* This flag is used to check if "-attrType" option is specified or not */
    int iMatches = 0; /* This flag is used to check if "-matches" option is specified or not */
    char * boolMatches = NULL;

    while ( (argIndex < argc) && (rc == 0) ) {

        if (isOption( argv[argIndex], list) ) {
            switch ( optVal->optionID ) {
            case ATTRTYPE:                                                 /* -attrType */
                iAttrType = 1;
                argIndex++;
                if (!argv[argIndex])
                    getAttrsUsage();
                attrsType = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the attrsType variable. */
                if ( attrsType == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                /* Check for the value of "-attrType"..it should be either of the following */

                if (strcmp(attrsType, "operational") != 0 && strcmp(attrsType, "language_tag") != 0
                        && strcmp(attrsType, "attribute_cache") != 0 && strcmp(attrsType, "unique") != 0
                        && strcmp(attrsType, "configuration") != 0) {
                    getAttrsUsage();
                }
                argIndex++;
                break;

            case MATCHES:                                                 /* -matches */
                iMatches = 1;
                argIndex++;
                if (!argv[argIndex])
                    getAttrsUsage();
                boolMatches = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the matches variable. */
                if (boolMatches == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                if (strcmp(boolMatches, "true") == 0) {
                    iMatch = LDAP_TRUE;
                } else if (strcmp(boolMatches, "false") == 0) {
                    iMatch = LDAP_FALSE;
                } else {
                    getAttrsUsage();
                }
                free(boolMatches);
                boolMatches = NULL;
                argIndex++;
                break;

            case OPTHELP:                                      /* help  */
                getAttrsUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                getAttrsUsage();
                rc = -1;
                break;

            default:
                break;
            }
        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            getAttrsUsage();
            rc = -1;
        }
    }

    if ( iAttrType == 0 || iMatches == 0 ) {
        getAttrsUsage();
    }

    getAttributes();

    return argIndex;
}

/********************/
/*-op evaluategroups*/
/********************/
/**********************************************************************
 * Function: decodeBerGroupEval
 * Purpose:  Function to decode the results from a group
 *           evaluation extended operation.
 * Inputs: 
 *        oidres - the oid returned from the extended operation
 *        result - the result returned from the extended operatin
 *        groupValues - place to store the sequence of group values
 *                      returned from the server.
 *
 * Pre-conditions: All inputs are properly initialized.  
 * Post-conditions:
 *    The groupValues will be set if the berval and oidres are valid.
 * Outputs: None.          
 ***********************************************************************/
void decodeBerGroupEval(const char * oidres,
                        struct berval * result,
                        char *** groupValues) /* Sequence of group values returned from the server */ {
    int response;  /* Stores the return code of the BER decoding function */
    int normalized; /*Flag that indicates if the groups are normalized*/

    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_GROUP_EVAL_RES_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {

        BerElement *ber = ber_init2(result);  /* Obtain the BerElement corresponding to the BER encoded format. */
        response = ber_scanf(ber, "{i{v}}", &normalized, groupValues);
        if (verbose)
            printf("ber_scanf(ber,\"{iv}\", normalized, groupValues)\n");

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }
    }
}

/**********************************************************************
 * Function: evaluategroups
 * Purpose:  Function to implement evaluategroups extended operation.
 *           The function generates the extended operation request, sends
 *           it to the server and calls the appropriate function to
 *           decode the results. The function then prints the results.
 * Inputs: None.
 *
 * Pre-conditions: entryDN and attrs are initialized.  
 *                 client is already bound to the server.
 * Post-conditions:
 *    The client is unbound.
 * Outputs: None.          
 ***********************************************************************/
void evaluategroups() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_GROUP_EVAL_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;
    int index = 0;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;

    char ** groupValues = NULL;

    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = ldap_create_group_eval_request(entryDn, attrs);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerGroupEval(oidres, valres, &groupValues);

        if (groupValues == NULL) {
            printf ( "\nNo values returned from the Server.\n");

        } else {
            printf("Group values are :\n");
            for (index = 0; groupValues != NULL && groupValues[index] != NULL; index++) {
                printf ( "%s \n", groupValues[index]);
                ldap_memfree(groupValues[index]);
            }
        }
    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

        rc = ldap_get_lderrno(ld, &dn, &msg);
        if ( msg != NULL && msg[0] != 0 )
            printf ( "msg: %s\n", msg);
    }

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);
    if (valres)
        ber_bvfree(valres);
}

/**********************************************************************
 * Function: parseGroupEvalInputs
 * Purpose:  This function parses the group evaluation inputs from the
 *           command line.  If any inputs are wrong or missing a usage
 *           will be printed.
 * Inputs:
 *   argIndex - index into the command line argument
 *   argc - number of arguments
 *   argv - command line inputs
 *   list - the list of options allowed by the evaluate groups operation
 *
 * Pre-conditions: all inputs are properly initialized.
 * Post-conditions:
 *    If successful the operation was completed.
 * Outputs: returns the incremented argIndex.
 *          
 ***********************************************************************/
int parseGroupEvalInputs(int argIndex, int argc, char *argv[], Option *list) {

    int iDN = 0; /* This flag is used to check if "-d" option is specified*/


    int index = 0;
    int counter = 0;

    while ( (argIndex < argc) && (rc == 0) ) {

        if (isOption( argv[argIndex], list) ) {
            switch ( optVal->optionID ) {
            case DN:                                                 /* -d */
                iDN = 1;
                argIndex++;
                if (!argv[argIndex])
                    groupEvalUsage();
                entryDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the entryDn variable. */
                if ( entryDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;
            case ATTRS:                                                 /* -a */
                /*determine the number of attrs specified*/
                index = ++argIndex;
                while (argv[index] != NULL) {
                    if (*(argv[index++]) == '-')
                        groupEvalUsage();
                    counter++;
                }

                /*Allocate the attrs list*/
                attrs = (char**)malloc((counter + 1) * sizeof(char*));
                if (attrs == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                if (counter > 0) {
                    index = 0;
                    while ((argv[argIndex] != NULL) && (index < counter)) {
                        attrs[index] = strdup (argv[argIndex++]);
                        if (attrs[index++] == NULL) {
                            fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                            exit(1);
                        }
                    }
                    attrs[index] = NULL;
                } else {
                    attrs[0] = NULL;
                }

                break;

            case OPTHELP:                                      /* help  */
                groupEvalUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                groupEvalUsage();
                rc = -1;
                break;

            default:
                break;
            }
        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            groupEvalUsage();
            rc = -1;
        }
    }

    if ( iDN == 0 ) {
        groupEvalUsage();
    }

    evaluategroups();

    return argIndex;
}

/*******************/
/* -op acctstatus  */
/*******************/
/**********************************************************************
 * Function: decodeBerGroupEval
 * Purpose:  Function to decode the results from a group
 *           evaluation extended operation.
 * Inputs: 
 *        oidres - the oid returned from the extended operation
 *        result - the result returned from the extended operation
 *
 * Pre-conditions: All inputs are properly initialized.  
 * Post-conditions: None
 *    
 * Outputs: None.          
 ***********************************************************************/
void decodeBerAcctStatus(const char * oidres,
                         struct berval * result) {
    int response;  /* Stores the return code of the BER decoding function */
    int status = -1; /*Flag that indicates the status of the account*/

    if (!oidres) {
        fprintf(stderr, "No extended response received\n");
        return;
    } else if (!result) {
        fprintf(stderr, "No extended response data received\n");
        return;
    } else if (strcmp(oidres, LDAP_ACCT_STATUS_RES_OID) != 0) {
        fprintf(stderr, "Unexpected extended response: %s\n", oidres);
        return;
    } else {

        BerElement *ber = ber_init2(result);  /* Obtain the BerElement corresponding to the BER encoded format. */
        response = ber_scanf(ber, "{i", &status);
        if (verbose)
            printf("ber_scanf(ber,\"{i\", status)\n");

        if (response == -1) {
            fprintf(stderr, "Ber_scanf failed");
            ber_free(ber, 1);
            return;
        }

        switch (status) {
        case ACCOUNT_OPEN:
            printf("\nAccount is open.\n");
            break;
        case ACCOUNT_LOCKED:
            printf("Account is locked.\n");
            break;
        case ACCOUNT_EXPIRED:
            printf("Account has expired.\n");
            break;
        default:
            fprintf(stderr, "Invalid result returned.\n");
            break;
        }
        ber_free(ber, 1);
    }
}
/**********************************************************************
 * Function: accountstatus
 * Purpose:  Function to implement acctstatus extended operation.
 *           The function queries the server to find the status of the account
 *           for the given DN, whether it is locked, expired or open
 * Inputs: None.
 *
 * Pre-conditions: entryDN is initialized.  
 *                 client is already bound to the server.
 * Post-conditions:
 *    The client is unbound.
 * Outputs: None.          
 ***********************************************************************/
void accountstatus() {
    struct berval *reqVal = NULL;
    const char *oid = LDAP_ACCT_STATUS_REQ_OID;
    char *oidres = NULL;
    struct berval *valres = NULL;
    int index = 0;

    /*used by ldap_get_lderrno*/
    char * dn = NULL;
    char * msg = NULL;


    /* Start a session with the server */
    initSession();

    /* Create the request */
    reqVal = ldap_create_account_status_request(entryDn);

    /* Issue the operation */
    rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);

    /* Check our result */
    if (rc == LDAP_SUCCESS) {
        decodeBerAcctStatus(oidres, valres);

    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));

        rc = ldap_get_lderrno(ld, &dn, &msg);
        if ( msg != NULL && msg[0] != 0 )
            printf ( "msg: %s\n", msg);
    }

    /* Unbind */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /*clean up*/
    if (reqVal)
        ber_bvfree(reqVal);
    if (oidres)
        ldap_memfree(oidres);

}

/**********************************************************************
 * Function: parseAccountStatusInputs
 * Purpose:  This function parses the account status inputs from the
 *           command line.  If any inputs are wrong or missing a usage
 *           will be printed.
 * Inputs:
 *   argIndex - index into the command line argument
 *   argc - number of arguments
 *   argv - command line inputs
 *   list - the list of options allowed by the account status operation
 *
 * Pre-conditions: all inputs are properly initialized.
 * Post-conditions:
 *    If successful the operation was completed.
 * Outputs: returns the incremented argIndex.
 *          
 ***********************************************************************/
int parseAccountStatusInputs(int argIndex, int argc, char *argv[], Option *list) {

    int iDN = 0; /* This flag is used to check if "-d" option is specified*/


    int index = 0;
    int counter = 0;

    while ( (argIndex < argc) && (rc == 0) ) {

        if (isOption( argv[argIndex], list) ) {
            switch ( optVal->optionID ) {
            case DN:                                                 /* -d */
                iDN = 1;
                argIndex++;
                if (!argv[argIndex])
                    acctStatusUsage();
                entryDn = strdup(argv[argIndex]);

                /* Make sure that we could allocate the memory for the entryDn variable. */
                if ( entryDn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                argIndex++;
                break;

            case OPTHELP:                                      /* help  */
                acctStatusUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                acctStatusUsage();
                rc = -1;
                break;

            default:
                break;
            }
        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            acctStatusUsage();
            rc = -1;
        }
    }

    if ( iDN == 0 ) {
        acctStatusUsage();
    }

    accountstatus();

    return argIndex;
}


/* Replication Error Log LDAP Extended Operation Functions */

/** Decode the server's response to the requested the replication error log
 * control extended operation.
 */
void decodeBerReplErrLog( const char *response_OID, struct berval *result ) {
    if ( !response_OID ) {
        fprintf(stderr, "No extended response received.\n");
        return;

    } else if ( !result ) {
        fprintf(stderr, "No extended response data received.\n");
        return;

    } else if ( strcmp(response_OID, LDAP_REPL_ERROR_LOG_CONTROL_RESPONSE_OID) != 0 ) {
        fprintf( stderr, "Unexpected extended response: %s.\n", response_OID );
        return;

    } else {
        BerElement *ber = NULL;
        int resultCode = 0;
        int response = 0;   /* return code of the BER decoding function */
        char *effected = NULL;
        char *msg = NULL;

        /* Obtain the BerElement corresponding to the BER encoded format. */
        ber = ber_init2( result );
        if ( ber == NULL ) {
            fprintf(stderr, "Unable to obtain BerElement from result.\n");
            return;
        }

        response = ber_scanf( ber, "{iaa", &resultCode, &effected, &msg );

        if (verbose) {
            printf("ber_scanf(ber,\"{iaa\", &resultCode, &effected, &msg)\n");

            if ( response != ( LBER_ERROR )) {
                printf("  resultCode=%d\n", resultCode );
                printf("  number of failures effected=%s\n", effected );
                printf("  msg=[%s]\n", msg );
            }
        } else if ( msg[1] == 'd' && msg[2] == 'n' && msg[3] == ':' ) {
            printf("%s", &msg[1] );
        }

        if ( response == ( LBER_ERROR )) {
            fprintf( stderr, "ber_scanf failed" );
            ber_free( ber, 1 );
            return;
        }

        switch ( resultCode ) {
        case LDAP_SUCCESS:
            printf("\nOperation completed successfully.\n");
            break;

        case LDAP_INSUFFICIENT_ACCESS:
            fprintf(stderr, "You do not have authority to perform this operation.\n");
            break;

        case LDAP_TIMEOUT:
            fprintf(stderr, "Operation did not complete within specified time.\n");
            break;

        default:
            printOtherError( resultCode, msg );
            break;
        }

        ldap_memfree( msg );
        ber_free( ber, 1 );
    }
}

/** Create an BER encoded request for the replication error log control
 * extended operation.
 */
struct berval * createRequestReplErrLog( LDAP_REPL_ERROR_LOG_OPTION option, unsigned long failure_ID ) {
    struct berval *request = NULL; /* Used for storing the encoded form to be returned to caller. */
    BerElement *ber = NULL;        /* Used for storing the encoded form to be converted to a berval. */
    char failure_ID_str[80];       /* Cannot encode long */

    if (verbose) {
        printf("ber_alloc(1)\n");
    }

    sprintf( failure_ID_str, "%lu", failure_ID );

    /* Allocate the memory for one sequence. */
    ber = ber_alloc_t( 1 );

    if ( ber == NULL ) {
        fprintf( stderr, "ber_alloc_t failed\n");

        return NULL;
    }

    if (verbose) {
        printf("ber_printf(ber, \"{iss}\", option, failure_ID, agreementDn )\n");
    }

    /* Compose the ber element here. */
    if (ber_printf(ber, "{iss}",
                   option,
                   failure_ID_str,
                   agreementDn ? agreementDn : "") == (LBER_ERROR)) {
        fprintf( stderr, "ber_printf failed\n");
        ber_free( ber, 1 );

        return NULL;
    }

    /* Encode the ber element into BER format. */
    if (ber_flatten( ber, &request ) == (LBER_ERROR)) {
        fprintf( stderr, "ber_flatten failed\n" );
        ber_free( ber, 1 );

        return NULL;
    }

    ber_free( ber, 1 );

    return request;
}

/** Process replication error log extended operations.
 */
void replErrorLog( LDAP_REPL_ERROR_LOG_OPTION option, unsigned long failure_ID ) {
    struct berval *reqVal = NULL;
    struct berval *result = NULL;
    const char *request_OID = LDAP_REPL_ERROR_LOG_CONTROL_REQUEST_OID;
    char *response_OID = NULL;
    char *dn  = NULL; /* Used by ldap_get_lderrno(). */
    char *msg = NULL; /* Used by ldap_get_lderrno(). */

    /* Start a session with the server */
    initSession();

    /* Create the request. */
    reqVal = createRequestReplErrLog( option, failure_ID );

    /* Send the request. */
    rc = ldap_extended_operation_s( ld, request_OID, reqVal, NULL, NULL, &response_OID, &result );

    /* Check the result. */
    if ( rc == LDAP_SUCCESS ) {
        decodeBerReplErrLog( response_OID, result );

    } else {
        fprintf( stderr, "\nldapexop: %d - %s \n", rc, ldap_err2string(rc));
    }

    rc = ldap_get_lderrno(ld, &dn, &msg);
    if ( msg != NULL && msg[0] != 0 ) {
        fprintf ( stderr, "msg: %s\n", msg);
    }

    /* Drop the connection to the server. */
    if ( tls )
        ldap_stop_tls_s_np ( ld );
    else
        ldap_unbind(ld);

    /* Clean up. */
    if ( reqVal ) {
        ber_bvfree( reqVal );
    }

    if ( response_OID ) {
        ldap_memfree( response_OID );
    }

    if (result) {
        ber_bvfree( result );
    }
}

/**
 * Parse all of the input parameters for the "controlreplerrlog" extended 
 * operation.
 */
int parseReplErrLogInputs(int argIndex, int argc, char *argv[], Option *list) {
    LDAP_REPL_ERROR_LOG_OPTION opt = LDAP_REPL_ERROR_LOG_UNSET;
    unsigned long failure_ID = 0;

    while ( (argIndex < argc) && (rc == 0) ) {
        if (isOption( argv[argIndex], list) ) {

            switch ( optVal->optionID ) {

            case REPLERRLOGSHOW:
                /* If an option was already specified, */
                if ( opt != LDAP_REPL_ERROR_LOG_UNSET ) {
                    /* Then display the usage. */
                    replErrLogUsage();
                }

                /* Set the option for the requested show operation. */
                opt = LDAP_REPL_ERROR_LOG_SHOW;

                argIndex++;
                /* If the value does not exist, */
                if ( argv[argIndex] == NULL ) {
                    /* Then display the usage. */
                    replErrLogUsage();
                }

                sscanf( argv[argIndex], "%lu", &failure_ID );

                argIndex++;
                break;

            case REPLERRLOGDELETE:
                /* If an option was already specified, */
                if ( opt != LDAP_REPL_ERROR_LOG_UNSET ) {
                    /* Then display the usage. */
                    replErrLogUsage();
                }

                /* Set the option for the requested delete operation. */
                opt = LDAP_REPL_ERROR_LOG_DELETE;

                argIndex++;
                /* If the value does not exist, */
                if ( argv[argIndex] == NULL ) {
                    /* Then display the usage. */
                    replErrLogUsage();
                }

                sscanf( argv[argIndex], "%lu", &failure_ID );

                argIndex++;
                break;

            case REPLERRLOGRETRY:
                /* If an option was already specified, */
                if ( opt != LDAP_REPL_ERROR_LOG_UNSET ) {
                    /* Then display the usage. */
                    replErrLogUsage();
                }

                /* Set the option for the requested retry operation. */
                opt = LDAP_REPL_ERROR_LOG_RETRY;
                argIndex++;
                /* If the value does not exist, */
                if ( argv[argIndex] == NULL ) {
                    /* Then display the usage. */
                    replErrLogUsage();
                }

                sscanf( argv[argIndex], "%lu", &failure_ID );

                argIndex++;
                break;

            case REPLERRLOGAGMTDN:                                          /* -ra */
                argIndex++;
                /* If the value does not exist, */
                if ( argv[argIndex] == NULL ) {
                    /* Then display the usage. */
                    replErrLogUsage();
                }

                agreementDn = strdup( argv[argIndex] );

                /* Make sure that we could allocate the memory for a copy of the DN. */
                if ( agreementDn == NULL) {
                    fprintf( stderr,
                             "Memory allocation failed near %d in %s; cannot continue.  Program terminated.\n",
                             __LINE__, __FILE__ );
                    exit(1);
                }

                argIndex++;
                break;

            case OPTHELP:                                      /* help  */
                replErrLogUsage();
                rc = -1;
                break;

            case OPTQUESTION:                                     /* ? */
                replErrLogUsage();
                rc = -1;
                break;

            default:
                break;
            }

        } else {
            /*
             * Unsupported option.
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex] );
            replErrLogUsage();
            rc = -1;
        }
    }

    if ( opt == LDAP_REPL_ERROR_LOG_UNSET ) {
        fprintf( stderr, "\nMust provide an option.\n");

        replErrLogUsage();
    }

    if (( opt == LDAP_REPL_ERROR_LOG_SHOW ) && ( failure_ID == 0 )) {
        fprintf( stderr,
                 "\nMust provide a specific failure ID with the -show option.\n");

        replErrLogUsage();
    }

    if ( agreementDn == NULL ) {
        fprintf( stderr,
                 "\nMust provide a replication agreement DN.\n");
        replErrLogUsage();
    }

    replErrorLog( opt, failure_ID );

    return argIndex;
}

/**
 *    This function handles all the input parameters for the "-op" operation
 **/
int parseExtendedOpInputs(int argIndex, int argc, char *argv[], Option *list) {
    int flag = 0;

    if (!argv[argIndex])
        usage();

    if (isOptionExtended( argv[argIndex], list) ) {

        switch ( optVal->optionID ) {

            /* Just required to uncomment the following lines to activate the
               option of full replication */
            /*     case EXOPTFULLREP:    */ /* fullRepl */
            /*     argIndex++;
                   argIndex = parseFullReplInputs(argIndex,argc,argv,fullReplList);
                   break;
            */
        case EXOPTQUIESCE:                                     /* quiesce */
            argIndex++;
            argIndex = parseQuiesceInputs(argIndex, argc, argv, quiesceList);
            break;

        case EXOPTCONTROLREP:                                     /* controlrepl */
            argIndex++;
            argIndex = parseControlReplInputs(argIndex, argc, argv, controlReplList);
            break;

        case EXOPTREPLERRLOG:                     /* controlreplerr */
            argIndex++;
            argIndex = parseReplErrLogInputs(argIndex, argc, argv, replErrLogList);
            break;

        case EXOPTCONTROLQUEUE:                                    /* controlqueue */
            argIndex++;
            argIndex = parseControlQueueInputs(argIndex, argc, argv, controlQueueList);
            break;

        case EXOPTCASCREP:                                     /* cascrepl */
            argIndex++;
            argIndex = parseCascadeInputs(argIndex, argc, argv, cascadeList);
            break;

        case EXOPTREADCONFG:                                     /* readconfig */
            argIndex++;
            argIndex = parseReadConfigInputs(argIndex, argc, argv, readConfigList);
            break;

        case EXOPTGETLOGSIZE:                                     /* getlogsize */
            argIndex++;
            argIndex = parseLogSizeInputs(argIndex, argc, argv, logSizeList);
            break;

        case EXOPTREADLOG:                                     /* readlog */
            argIndex++;
            argIndex = parseReadLogInputs(argIndex, argc, argv, readLogList);
            break;

        case EXOPTCLEARLOG:                                     /*  clearlog */
            argIndex++;
            argIndex = parseClearLogInputs(argIndex, argc, argv, clearLogList);
            break;

        case EXOPTUNBIND:                                   /* unbind */
            argIndex++;
            argIndex = parseUnbindInputs(argIndex, argc, argv, unbindList);
            break;

        case EXOPTATTRTYPE:                                     /*  uniqueattr */
            argIndex++;
            argIndex = parseUniqueAttrInputs(argIndex, argc, argv, uniqueAttrList);
            break;

        case EXOPTGETATTRTYPE:                                     /*  getattributes */
            argIndex++;
            argIndex = parseGetAttrsInputs(argIndex, argc, argv, getAttrList);
            break;

        case EXOPTGETUSRTYPE:                                     /*  getusertype */
            argIndex++;
            if (!argv[argIndex]) {
                getUserType();
            } else
                getUserTypeUsage();
            break;

        case EXOPTSTOPSERVER:                                     /*  stopserver */
            argIndex++;
            if (!argv[argIndex]) {
                stopServer();
            } else
                stopServerUsage();
            break;

        case EXOPTGROUPEVAL:                                     /*  groupEvaluation */
            argIndex++;
            argIndex = parseGroupEvalInputs(argIndex, argc, argv, evalGroupsList);
            break;

        case EXOPTREPLTOPOLOGY:                                   /* repltopology */
            argIndex++;
            argIndex = parseReplTopologyInputs(argIndex, argc, argv, replTopologyList);
            break;

        case EXOPTACCTSTATUS:                                      /* acctstatus */
            argIndex++;
            argIndex = parseAccountStatusInputs(argIndex, argc, argv, acctStatusList);
            break;

        default:
            break;
        }
    } else {
        /*
         * Unsupported option
         */
        fprintf( stderr, "\nUnsupported Extended Operation option: %s.\n", argv[argIndex] );
        usage();
        rc = -1;
    }
    return argIndex;
}


int main( int argc, char *argv[] ) {
    int noInputFlag = 0;
    int extendedOperationFlag = 0;
    LDAPVersion ver;     /* This structure is basically to store the LDAP version info */
    char * tmpPort = NULL;
    char * tmpDebug = NULL;

    /*
     * Process command-line arguments
     */

    while ( (argIndex < argc) && (rc == 0) ) {
        noInputFlag = 1;

        if (isOption( argv[argIndex], optList) ) {
            switch ( optVal->optionID ) {
            case VERBOSE:                                      /* -v */
                verbose = LDAP_TRUE;
                argIndex++;
                break;

            case HOST:                                                  /* -h */
                argIndex++;
                if (!argv[argIndex])
                    usage();
                host = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the host variable. */
                if ( host == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                break;

            case PORT:                                      /* -p */
                argIndex++;
                if (!argv[argIndex])
                    usage();
                if (!is_num(argv[argIndex])) {
                    fprintf(stderr, "\nIncorrect ldap port.\n");
                    usage();
                }
                tmpPort = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the tmpPort variable. */
                if ( tmpPort == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                port = atoi(tmpPort);
                free(tmpPort);

                break;

            case DN:                                                    /* -D */
                argIndex++;
                if (!argv[argIndex])
                    usage();
                dn = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the dn variable. */
                if ( dn == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                break;

            case PASSWD:                                      /* -w */
                argIndex++;
                if (!argv[argIndex])
                    usage();

                passwd = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the passwd variable. */
                if ( passwd == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                if (strcmp(passwd, "?") == 0) {
                    /* Since the password is not to be specified in the visible mode,
                       call the routine to get the password in the invisible mode */

                    passwd = getPassword();
                }
                break;

            case PROXY:                                      /* -y */
                argIndex++;
                if (!argv[argIndex])
                    usage();

                proxydn = strdup( argv[argIndex++] );
                /*  rc = ldap_proxy_dn_prefix(&proxydn,argv[argIndex++]);  */
                if (( proxydn == NULL) || (rc != LDAP_SUCCESS)) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                stringLen = strlen(proxydn);
                rc = ldap_add_control (LDAP_PROXY_CONTROL_OID, stringLen,
                                       proxydn, 1, &opControls);
                free(proxydn);
                if (rc != LDAP_SUCCESS) {
                    fprintf( stderr, "proxy control creation failed; cannot continue, Program terminate. \n");
                    exit(1);
                }
                break;
            case FIPS:                                                   /* -x */
                use_FIPS_mode = LDAP_TRUE;
                argIndex++;
                break;

            case SSL:                                                   /* -Z */
                if (tls) {
                    fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
                    usage();
                }
                ssl = LDAP_TRUE;
                argIndex++;
                break;

            case TLS:                                                   /* -Y */
                if (ssl) {
                    fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
                    usage();
                }
                tls = LDAP_TRUE;
                argIndex++;
                break;

            case KEYFILE:                                      /* -K */
                keys = LDAP_TRUE;
                argIndex++;
                if (!argv[argIndex])
                    usage();
                keyFile = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the keyFile variable. */
                if ( keyFile == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                break;

            case KEYNAME:                                      /* -N */
                argIndex++;
                if (!argv[argIndex])
                    usage();
                keyName = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the keyName variable. */
                if ( keyName == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                break;

            case KEYPW:                                      /* -P */

                argIndex++;
                if (!argv[argIndex])
                    usage();
                keyPwd = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the keyPwd variable. */
                if ( keyPwd == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                break;

            case MECHANISM:                                     /* -m */

                argIndex++;
                if (!argv[argIndex])
                    usage();
                mechanism = strdup( argv[argIndex++]);

                /* Make sure that we could allocate the memory for the mechanism variable. */
                if ( mechanism == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                break;

            case CHARSET:                                      /* -C */

                argIndex++;
                if (!argv[argIndex])
                    usage();
                charset = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the charset variable. */
                if ( charset == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }
                break;


            case VERSION:                                      /* -e */
                argIndex++;
                ldap_version(&ver);
                printf ("\nSDK Version:     \t%d\n", ver.sdk_version);
                printf ("Protocol Version:\t%d\n", ver.protocol_version);
                printf ("SDK Build Level: \t%s\n", ver.sdk_build_level);
                exit(0);

            case LEVEL:                                      /* -d */

                argIndex++;
                if (!argv[argIndex])
                    usage();

                tmpDebug = strdup( argv[argIndex++] );

                /* Make sure that we could allocate the memory for the tmpDebug variable. */
                if ( tmpDebug == NULL) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit(1);
                }

                debugLevel = atoi(tmpDebug);
                free(tmpDebug);
                break;

            case DIGEST_USERNAME:       /* -U */
                argIndex++;
                if (!argv[argIndex])
                    usage();

                digest_username = strdup( argv[argIndex++] );

                if ( digest_username == NULL ) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n");
                    exit( -1 );
                }
                break;

            case DIGEST_REALM:       /* -G */
                argIndex++;
                if (!argv[argIndex])
                    usage();

                digest_realm = strdup( argv[argIndex++] );

                if ( digest_realm == NULL ) {
                    fprintf( stderr, "Memory allocation failed; cannot continue.  Program terminated.\n" );
                    exit( -1 );
                }
                break;

            case OPTHELP:                                      /* help  */

                usage();
                argIndex++;
                break;

            case OPTQUESTION:                                     /* ? */

                usage();
                argIndex++;
                break;


            case OPTEXTENDEDOP:                                    /* -op */

                argIndex++;
                if (!argv[argIndex])
                    usage();
                extendedOperationFlag = 1;
                argIndex = parseExtendedOpInputs(argIndex, argc, argv, extendedOptList);

                break;
            default:
#ifdef _AIX

                if ( i == '\377' )
                    ;
#endif

                usage();
                break;
            }
        } else {
            /*
             * Unsupported option
             */
            fprintf( stderr, "\nUnsupported option: %s.\n", argv[argIndex]);
            usage();
            rc = -1;
        }
    }

    /* If no input is specified then display the usage */
    if (noInputFlag == 0 || extendedOperationFlag == 0)
        usage();

    return rc;
}

/* This function is used to check if the specified option is valid or not
 * This function handles the options starting with "-"
 */

int isOption( char *argument, Option *optionList) {
    Option *cursor = optionList;
    int found = LDAP_FALSE;

    if ( *argument == '-' ) {
        while ( (cursor->optionName != NULL) && !found ) {

            if ( strlen(argument) - 1 >= (unsigned int)cursor->optionMinAbbrev &&
                    strlen(argument) - 1 <= strlen(cursor->optionName) ) {
                if ( strncmp( argument + 1, cursor->optionName, max( strlen(argument) - 1, (unsigned int)cursor->optionMinAbbrev) ) == 0 ) {
                    found = LDAP_TRUE;
                }
            }

            if (!found) {
                cursor++;
            }
        }
    }

    if ( !found ) {
        optVal = NULL;
    } else {

        optVal = cursor;
    }

    return (found);
}


/* This function is used to check if the specified option is valid or not
 * This function handles the options which are not suppose to start with "-"
 */

int isOptionExtended( char *argument, Option *optionList) {
    Option *cursor = optionList;
    int found = LDAP_FALSE;

    while ( (cursor->optionName != NULL) && !found ) {

        if ( strlen(argument) >= (unsigned int)cursor->optionMinAbbrev &&
                strlen(argument) <= strlen(cursor->optionName) ) {
            if ( strncmp( argument, cursor->optionName, max( strlen(argument), (unsigned int)cursor->optionMinAbbrev) ) == 0 ) {
                found = LDAP_TRUE;
            }
        }

        if (!found) {
            cursor++;
        }
    }

    if ( !found ) {
        optVal = NULL;
    } else {

        optVal = cursor;
    }

    return (found);
}

/* end of ldapexop.c */
