/*
 * COMPONENT_NAME: ldap.clients
 *
 * ABSTRACT: generic program to modify or add entries using LDAP
 *
 * ORIGINS: 202,27
 * CHANGE HISTORY:
 *
 * DATE        INITIALS  DEFECT   COMMENT
 * ----        --------  ------   -------
 * 01/27/06    alpesh    D92899   ldapmodify allows invalid inputs
 *
*/ 
/*
 * Copyright (c) 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/* ldapmodify.c - generic program to modify or add entries using LDAP */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#if !defined( WIN32 )
#include <sys/file.h>
#endif
#include <fcntl.h>
#if !defined(WIN32)
#include <unistd.h>
#endif
#include <ldap.h>
#include <ldapssl.h>
#include <ldif.h>

#define LDAPMODIFY_REPLACE 1
#define LDAPMODIFY_ADD  2

#define DEFHOST "localhost"
#define CRYPTOSYNC_ATTR_NAME "ibm-slapdCryptoSync"
#define CRYPTOSALT_ATTR_NAME "ibm-slapdCryptoSalt"

#if defined( WIN32 ) || defined( _WIN32 )
#define strcasecmp stricmp
#endif

#define safe_realloc( ptr, size ) ( ptr == NULL ? malloc( size ) : \
      realloc( ptr, size ))

#define LDAPMOD_MAXLINE  4096

/* strings found in replog/LDIF entries (mostly lifted from slurpd/slurp.h) */
#define T_REPLICA_STR  "replica"
#define T_DN_STR   "dn"
#define T_CHANGETYPESTR  "changetype"
#define T_ADDCTSTR   "add"
#define T_MODIFYCTSTR  "modify"
#define T_DELETECTSTR  "delete"
#define T_MODRDNCTSTR  "modrdn"
#define T_MODDNCTSTR   "moddn"
#define T_MODOPADDSTR  "add"
#define T_MODOPREPLACESTR "replace"
#define T_MODOPDELETESTR "delete"
#define T_MODSEPSTR   "-"
#define T_NEWRDNSTR   "newrdn"
#define T_DELETEOLDRDNSTR "deleteoldrdn"
#define T_NEWSUPERIORSTR  "newsuperior"
#define T_CONTROLSTR "control"

#ifdef NEEDPROTOS
static void freepmods( LDAPMod **pmods );
#endif /* NEEDPROTOS */

/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else 
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( void );

const char* ssl_err2string( int );

#ifndef _WIN32
#ifdef _GCC3
#include <errno.h>
#else
extern int errno;
#endif
#endif

/* global variables */
static char *prog;
static char *binddn = NULL;
static char *passwd = (char *) "";
static char *ldaphost = (char *)DEFHOST;
static char *mech = NULL;
static int badFile = 0;
static int hoplimit = 10;
static int ldapport = LDAP_PORT;
static int verbose = 0;
static int contoper = 0;
static int reject = 0;
static int doit = 1;
static int valsfromfiles = 0;
static int operation = LDAPMODIFY_REPLACE;
static int referrals = LDAP_OPT_ON;
static int deref = LDAP_DEREF_NEVER;
static int ldapversion = LDAP_VERSION3;
static int DebugLevel = 0;
static FILE *fp = NULL;
static FILE *rfp = NULL;
static char *charset = NULL;
static int keys = 0;
static int tls = 0;
static int ssl = 0;
static int use_FIPS_mode = 0; /* 0=off 1=on */
static char *keyfile = NULL;
static char *keyfile_pw = NULL;
static char *keyfile_label = NULL;
static char *debug_ptr = NULL;
static int ignoreSpaces = 1;
static int doNotReplicate = 0;
static char * pseudoCryptoSync = NULL;
static char * pseudoCryptoSalt = NULL;
static int pseudoCryptoSyncEntry = 0;

/*Controls*/
LDAPControl** bindControls = NULL; /*Controls sent on the bind*/
LDAPControl** opControls = NULL; /*The base set of controls for the modify*/
LDAPControl** serverControls = NULL; /*The base controls plus any controls from the ldif file*/

/* Digest-MD5 */
static char *digest_username = NULL;
static char *digest_realm = NULL;

/*
** Display usage statement and exit.
*/
void usage() {
   fprintf(stderr, "\nSends modify or add requests to an LDAP server.\n");
   fprintf(stderr, "usage:\n");
   fprintf(stderr, "    %s [options] [-i file]\n", prog);
   fprintf(stderr, "where:\n");
   fprintf(stderr, "    file: name of input file\n");
   fprintf(stderr, "note:\n");
   fprintf(stderr, "    standard input is used if file is not specified\n");
   fprintf(stderr, "options:\n" );
   fprintf(stderr, "    -a           force add operation as default\n");
   fprintf(stderr, "    -b           support binary values from files (old style paths)\n");
   fprintf(stderr, "    -c           continuous operation; do not stop processing on error\n");
   fprintf(stderr, "    -C charset   character set name to use, as registered with IANA\n");
   fprintf(stderr, "    -d level     set debug level in LDAP library\n");
   fprintf(stderr, "    -D dn        bind dn\n");
   fprintf(stderr, "    -e errorfile specifies the file to which rejected entries\n");
   fprintf(stderr, "                 will be written, requires the -c option\n");
   fprintf(stderr, "    -f file      read entry modification information from specified file\n");
   fprintf(stderr, "    -g           do not strip trailing spaces on attribute values\n");
   fprintf(stderr, "    -G realm     realm used for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -h host      LDAP server host name\n");
   fprintf(stderr, "    -i file      read entry modification information from specified file\n");
   fprintf(stderr, "    -k           use server administration control\n");
   fprintf(stderr, "    -K keyfile   file to use for keys\n");
   fprintf(stderr, "    -l           do not replicate the modified entry\n");
   fprintf(stderr, "    -m mechanism perform SASL bind with the given mechanism\n");
   fprintf(stderr, "    -n           show in standard output changes being\n");
   fprintf(stderr, "                 made without perfoming action on directory\n");
   fprintf(stderr, "    -M           Manage referral objects as normal entries.\n");
   fprintf(stderr, "    -N key_name  private key name to use in keyfile\n");
   fprintf(stderr, "    -O maxhops   maximum number of referrals to follow in a sequence\n");
   fprintf(stderr, "    -p port      LDAP server port number\n");
   fprintf(stderr, "    -P key_pw    keyfile password\n");
   fprintf(stderr, "    -r           force replace operation as default\n");
   fprintf(stderr, "    -R           do not chase referrals\n");
   fprintf(stderr, "    -U username  user name for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -v           verbose mode\n");
   fprintf(stderr, "    -V version   LDAP protocol version (2 or 3; default is 3)\n");
   fprintf(stderr, "    -w password  bind password or '?' for non-echoed prompt\n");
   fprintf(stderr, "                 use backslash '\\?' to avoid matching single character filenames (UNIX only)\n" );
   fprintf(stderr, "    -x           use FIPS mode processing (SSL/TLS only)\n");
   fprintf(stderr, "    -y proxydn   set proxied id for proxied authorization operation\n");
   fprintf(stderr, "    -Y           use a secure LDAP connection (TLS)\n");
   fprintf(stderr, "    -Z           use a secure LDAP connection (SSL)\n");
   exit(1);
}


/*
**      Print a memory error message and exit.
*/
void mem_error(const char *varName) {
   fprintf(stderr, "%s: Memory allocation error occured.\n", varName);
   exit(1);
}

/*
** Parse command line arguments.
*/
void parse_arguments(int argc, char **argv) {
   int i, port = 0;
   const char *optpattern = "abcRMxYZnrvkgl?h:V:p:D:w:d:i:f:m:O:K:P:N:C:G:y:U:e:";
   extern char *optarg;
   extern int optind;
   int manageDsa = LDAP_FALSE;
   int adminControl = LDAP_FALSE;
   int Proxy_control_requested = LDAP_FALSE;
   int rc = LDAP_SUCCESS;
   char *proxyDN;
   int stringLen = 0;

   fp = stdin;
   while ((i = getopt(argc, argv, optpattern)) != EOF) {
      switch ( i ) {
      case 'V':        /* version */
         ldapversion = atoi(optarg);
         if ((ldapversion != LDAP_VERSION2) && (ldapversion != LDAP_VERSION3)) {
            fprintf(stderr, "Unsupported version level supplied.\n");
            usage();
         }
         break ;
      case 'O':               /* set maximum referral hop count  */
         hoplimit = atoi( optarg );
         break;
      case 'm':               /* use SASL bind mechanism  */
         if (!(mech = strdup ( optarg )))
            mem_error("mech");
         break;
      case 'a':         /* add */
         operation = LDAPMODIFY_ADD;
         break;
      case 'b':         /* read values from files (for binary attributes) */
         valsfromfiles = 1;
         break;
      case 'c':         /* continuous operation */
         contoper = 1;
         break;
      case 'h':         /* ldap host */
         if (!(ldaphost = strdup( optarg )))
            mem_error("host");
         break;
      case 'D':         /* bind DN */
         if (!(binddn = strdup( optarg )))
            mem_error("bindDN");
         break;
      case 'w':         /* password */
         if (optarg && optarg[0] == '?' && ( strlen(optarg) == 1 )) {
            passwd = getPassword();

         } else
            if (!(passwd = strdup( optarg )))
               mem_error("password");
         break;
      case 'd':        /* debug */
         DebugLevel = atoi(optarg);
         break;
      case 'i':
      case 'f':         /* read from file */
         if ((optarg[0] == '-') && (optarg[1] == '\0'))
            fp = stdin;
         else if ((fp = fopen( optarg, "r" )) == NULL) {
            perror( optarg );
            if (contoper == 1)
               fp = stdin;
            exit( 1 );
         }
         break;
      case 'e':    /* write to reject file */
         if ( (optarg == NULL) || ((rfp = fopen(optarg, "w")) == NULL) ) {
            perror(optarg);
            exit(1);
         }
         reject = 1;
         break;
      case 'n':    /* print changes, don't actually do them */
         doit = 0;
         break;
      case 'g':        /* read from file */
         ignoreSpaces = 0;
         break;
      case 'p':        /* port */
         ldapport = atoi( optarg );
         port = 1;
         break;
      case 'r':         /* default is to replace rather than add values */
         operation = LDAPMODIFY_REPLACE;
         break;
      case 'R':          /* don't automatically chase referrals */
         referrals = LDAP_OPT_OFF;
         break;
      case 'l':          /* don't replicate the entry after it's modified */
         doNotReplicate = 1;
         if ((rc = ldap_add_control (LDAP_REPL_DO_NOT_REPLICATE_CONTROL_OID, 0 ,
                                     NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'M':         /* manage referral objects as normal entries */
         manageDsa = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &bindControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }

         break;
      case 'v':         /* verbose mode */
         verbose++;
         break;
      case 'K':        /* keyfile */
         if (!(keyfile = strdup( optarg )))
            mem_error("keyfile name");
         keys = 1;
         break;
      case 'P':        /* keyfile password */
         if (!(keyfile_pw = strdup( optarg )))
            mem_error("keyfile password");
         break;
      case 'N':        /* keyfile label */
         if (!(keyfile_label = strdup( optarg )))
            mem_error("keyfile label");
         break;
      case 'Y':       /* TLS */
         if (ssl) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         tls = 1;
         break;
      case 'x':
         use_FIPS_mode = 1;
         break;
      case 'Z':        /* SSL */
         if (tls) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         ssl = 1;
         break;
      case 'C':        /* charset */
         if (!(charset = strdup( optarg )))
            mem_error("charset");
         break;
      case 'k':      /* admin control */
         adminControl = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_SERVER_ADMIN_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &bindControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         if ((rc = ldap_add_control (LDAP_SERVER_ADMIN_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'y':     /* proxied authorization control */
         rc = ldap_proxy_dn_prefix(&proxyDN, optarg);
         if (rc == LDAP_SUCCESS) {
            stringLen = strlen(proxyDN);
            rc = ldap_add_control (LDAP_PROXY_CONTROL_OID, stringLen,
                                   proxyDN, 1, &opControls);
            free(proxyDN);
         }
         if (rc != LDAP_SUCCESS) {
            fprintf(stderr, "Incorrect Proxied DN specified. \n");
            usage();
         } else
            Proxy_control_requested = LDAP_TRUE;
         break;
      case 'U':     /* DIGEST-MD5 Username */
         if ( !(digest_username = strdup( optarg )) ) {
            mem_error( "Username" );
         }
         break;
      case 'G':     /* DIGEST-MD5 Realm */
         if ( !(digest_realm = strdup( optarg )) ) {
            mem_error( "realm" );
         }
         break;
      case '?':
      default:
         usage();
      }
   }

   if ((! (tls || ssl)) && keys) /* Specified a key file, but no security protocol */
      ssl = 1;                  /* Use SSL by default */

   if ( (manageDsa | adminControl | Proxy_control_requested) && ( ldapversion == LDAP_VERSION2 ) ) {
      fprintf( stderr, "\n-M and -k options require at least version 3. ");
      if (manageDsa)
         fprintf(stderr, "-M ignored. ");
      if (adminControl)
         fprintf(stderr, "-k ignored.");
      fprintf(stderr, "\n\n");
   }

   if ((reject == 1) && (contoper == 0)) {
      fprintf(stderr, "You must specify continuous operation with the reject file option\n");
      usage();
   }

   if (argc - optind != 0)
      usage();

   /* Use default SSL port if none specified */
   if (( port == 0 ) && ( ssl ))
      ldapport = LDAPS_PORT;

   if ( ! DebugLevel )
      if ( ( debug_ptr = getenv ( "LDAP_DEBUG" ) ) )
         DebugLevel = atoi ( debug_ptr );

   /*Set up the controls*/
   /*Add the password policy controls*/
   if (ldapversion == LDAP_VERSION3) {
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &bindControls)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &opControls)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }
   }
}

/*  Get a password from the user but don't display it. */
char* getPassword( void ) {
   char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

   char in = '\0';                  /* Input character */
   int len = 0;                    /* Length of password */
#else

   struct termios echo_control;
   struct termios save_control;

   int fd = 0;                 /* File descriptor */
   int attrSet = 0;            /* Checked later for reset */

   /* Get the file descriptor associated with stdin. */
   fd = fileno( stdin );

   if (tcgetattr( fd, &echo_control ) != -1) {
      save_control = echo_control;
      echo_control.c_lflag &= ~( ECHO | ECHONL );

      if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
         fprintf(stderr, "Internal error setting terminal attribute.\n");
         exit( errno );
      }

      attrSet = 1;
   }
#endif

   /* Prompt for a password. */
   fputs( "Enter password ==> ", stdout );
   fflush( stdout );

#ifdef _WIN32
   /* Windows 9x/NT will always read from the console, i.e.,
      piped or redirected input will be ignored. */
   while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
      in = _getch();

      if (in != '\r') {
         supplied_password[len] = in;
         len++;

      } else {
         supplied_password[len] = '\0';
      }
   }
#else
   /* Get the password from stdin. */
   fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

   /* Remove the newline at the end. */
   supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
   /* Reset the terminal. */
   if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
      fprintf(stderr, "Unable to reset the display.\n");
   }
#endif

   fprintf( stdout, "\n" );

   return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}

/*
** Rebind callback function.
*/
int rebindproc(LDAP *ld, char **dnp, char **pwp, int *methodp, int freeit) {
   if ( !freeit ) {
      if ( mech != NULL)
         *methodp = LDAP_AUTH_SASL;
      else
         *methodp = LDAP_AUTH_SIMPLE;
      if ( binddn != NULL ) {
         *dnp = strdup( binddn );
         *pwp = strdup ( passwd );
      } else {
         *dnp = NULL;
         *pwp = NULL;
      }
   } else {
      free ( *dnp );
      free ( *pwp );
   }
   return ( LDAP_SUCCESS );
}

/*
** Connect and bind to server.
*/
LDAP *connect_to_server() {
   LDAP *ld;
   int failureReasonCode, rc, authmethod;
   struct berval ber;
   struct berval *server_creds;
   int msgidp = 0;
   int errcode = 0;
   int error = 0;
   int rc2 = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *errmsg = NULL;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;
   LDAPControl **bind_client_controls = NULL;

   /* Set FIPS mode processing if requested for TLS and SSL. */
   if (( tls == 1 || ssl == 1 ) && use_FIPS_mode == 1 ) {
      if (verbose)
         printf( "ldap_ssl_set_fips_mode_np(1)\n");

      rc = ldap_ssl_set_fips_mode_np( 1 );
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_set_fips_mode_np failed\n" );
         exit( rc ) ;
      }
   }

   /* call ldap_ssl_client_init if V3 and SSL */
   if ((tls || ssl) && (ldapversion == LDAP_VERSION3)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }

      if (verbose)
         printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"));

      rc = ldap_ssl_client_init( keyfile, keyfile_pw, 0, &failureReasonCode ) ;
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                  rc, failureReasonCode );
         fprintf( stderr, "%s\n", ssl_err2string( failureReasonCode ));
         exit( 1 ) ;
      }
   }

   /* open connection to server */
   if (ldapversion == LDAP_VERSION3) {
      if (ssl) {
         if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport,
                   ((keyfile_label) ? keyfile_label : "NULL"));
         ld = ldap_ssl_init( ldaphost, ldapport, keyfile_label ) ;
         if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed\n" ) ;
            perror( "ldap_ssl_init" ) ;
            exit( 1 ) ;
         }
      } else {
         if (verbose)
            printf("ldap_init(%s, %d) \n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport);
         if ((ld = ldap_init(ldaphost, ldapport)) == NULL) {
            perror( "ldap_init" );
            exit(1);
         }
      }
   } else {
      if ( verbose )
         printf("ldap_open( %s, %d )\n",
                ((ldaphost) ? ldaphost : "NULL"), ldapport );

      if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
   }

   /* set options */
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
   if (ldapversion == LDAP_VERSION3) {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )&deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)&hoplimit);
      ldap_set_option( ld, LDAP_OPT_REBIND_FN, (void *)rebindproc);
   } else {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )(long)DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )(long)deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)(long)hoplimit);
   }
   /* don't chase referrals during the bind */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )LDAP_OPT_OFF);
   if (binddn != NULL)
      ldap_set_rebind_proc( ld, (LDAPRebindProc)rebindproc );
   if (charset != NULL) {
      if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
         fprintf(stderr, "unsupported charset %s\n", charset);
         exit(0);
      }
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
   }

   /* call ldap_ssl_start if V2 and SSL */
   if ( (tls || ssl) && (ldapversion == LDAP_VERSION2)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }
      if (verbose)
         printf( "ldap_ssl_start( ld, %s, %s, %s )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"),
                 ((keyfile_label) ? keyfile_label : "NULL"));

      rc = ldap_ssl_start(ld, keyfile, keyfile_pw, keyfile_label);
      if ( rc != 0) {
         fprintf( stderr, "ldap_ssl_start failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr, "ldap_ssl_start: %s \n", errormsg);
         exit( 1 );
      }
   }

   if (tls) {                   /* Start the TLS session */
      if ((rc = ldap_start_tls_s_np(ld, NULL)) != LDAP_SUCCESS) {
         fprintf(stderr, "ldap_start_tls_s() failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr, "ldap_start_tls_s_np: %s\n", errormsg);
         exit(rc);
      }
   }

   /* bind to server */
   if (ldapversion == LDAP_VERSION3) {
      if ( ! mech ) /* Use simple bind */
      {
         ber.bv_len = strlen ( passwd );
         ber.bv_val = passwd;
         rc = ldap_sasl_bind(ld, binddn, LDAP_SASL_SIMPLE, &ber, bindControls, NULL, &msgidp);
         rc2 = rc; /* Store this result code for later user. */
         if (rc == LDAP_SUCCESS) {
            error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
            rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                                    NULL, &returnedControls, 1 );
         }
      }
      else /* Presence of mechanism means SASL bind */ {
         /*
         ** Special case for mech="EXTERNAL".  Unconditionally set bind DN
         ** and credentials to NULL.  This option should be used in tandem
         ** with SSL and client authentication.  For other SASL mechanisms,
         ** use the specified bind DN and credentials.
         */
         if (strcmp(mech, LDAP_MECHANISM_EXTERNAL) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, bindControls, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr, "ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr, "ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else if (strcmp(mech, LDAP_MECHANISM_GSSAPI) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, bindControls, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr, "ldap_sasl_bind_s: %s \n", errormsg);
               ldap_get_option( ld, LDAP_OPT_EXT_GSS_ERR, &failureReasonCode);
               fprintf(stderr, "ldap_sasl_bind_s: extended error code: Major Status: %i\n", failureReasonCode);
               fprintf(stderr, "ldap_sasl_bind_s: extended error code: Minor Status: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else /* other SASL mechanisms */ {
            if (strcmp(mech, LDAP_MECHANISM_DIGEST_MD5) == 0) {
               if (digest_username) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_USER_NAME_OID, strlen(digest_username) , digest_username, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
               if (digest_realm) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_REALM_NAME_OID, strlen(digest_realm) , digest_realm, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
            }

            ber.bv_len = strlen ( passwd );
            ber.bv_val = passwd;
            rc = ldap_sasl_bind_s (ld, binddn, mech, &ber,
                                   bindControls, bind_client_controls,
                                   &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr, "ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr, "ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         }

         /* Get any controls sent by the server during the sasl_bind_s */
         rc2 = rc;

         if ( rc == LDAP_SUCCESS ) {
            rc = ldap_get_bind_controls( ld, &returnedControls );
         }
      }

      /*Free the client controls*/
      ldap_controls_free(bind_client_controls);

      /* moved error reporting to common code */
      if (rc == LDAP_SUCCESS) {
         if (returnedControls != NULL) {
            ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
         }
      }
      if (rc2 != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc2);
      } else if (rc != LDAP_SUCCESS) {
         errormsg = ldap_err2string(error);
      } else if (errcode != LDAP_SUCCESS) {
         errormsg = ldap_err2string(errcode);
      }
      if (controlerr) {
         errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
      }
      if (controlwarn) {
         warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
      }
      if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
         fprintf(stderr, "ldap_simple_bind:  ");
         if (errormsg != NULL) {
            fprintf(stderr, errormsg);
            fprintf(stderr, " --- ");
         }
         if (errormsgctrl != NULL) {
            fprintf(stderr, errormsgctrl);
            fprintf(stderr, "\n");
         }
         if (warnmsgctrl != NULL) {
            fprintf(stderr, warnmsgctrl, controlres);
            fprintf(stderr, "\n");
         }
      } else if (errormsg != NULL) {
         fprintf(stderr, "ldap_simple_bind: %s \n", errormsg);
      }
      if (rc2 != LDAP_SUCCESS) {
         /*
         ** ldap_get_exterror only valuable for ssl communication.
         ** In this example, for LDAP v3, the bind is the first
         ** instance in which communication actually flows to the
         ** server.  So, if there is an ssl configuration error or
         ** other ssl problem, this will be the first instance where
         ** it will be detected.
         */
         if (ssl) {
            fprintf( stderr, "Attempted communication over SSL.\n");
            fprintf( stderr, "  The extended error is %d.\n", ldap_get_exterror(ld));
         }
         exit( rc2 );
      } else if (errcode != LDAP_SUCCESS) {
         exit( errcode );
      }

   } else /* LDAP V2 */ {
      authmethod = LDAP_AUTH_SIMPLE;

      rc = ldap_bind_s( ld, binddn, passwd, authmethod );

      if ( rc != LDAP_SUCCESS ) {
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr, "ldap_bind_s: %s \n", errormsg);
         exit( rc );
      }
   }

   /* during the operation, use user setting to chase referrals */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )(long)referrals);
   return ld;
}

/* ssl_err2string returns a string describing the error associated with
** the given error code from an SSL related call.
*/
const char * ssl_err2string( int ssl_error ) {
   switch ( ssl_error ) {
   case LDAP_SSL_EXTENDED_OK:
      return LDAP_SSL_EXTENDED_OK_STR;
   case LDAP_SSL_EXTENDED_INVALID_HANDLE:
      return LDAP_SSL_EXTENDED_INVALID_HANDLE_STR;
   case LDAP_SSL_EXTENDED_API_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_API_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_INTERNAL_ERROR:
      return LDAP_SSL_EXTENDED_INTERNAL_ERROR_STR;
   case LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE:
      return LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE_STR;
   case LDAP_SSL_EXTENDED_INVALID_STATE:
      return LDAP_SSL_EXTENDED_INVALID_STATE_STR;
   case LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND:
      return LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND_STR;
   case LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION:
      return LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_CRYPTO:
      return LDAP_SSL_EXTENDED_ERROR_CRYPTO_STR;
   case LDAP_SSL_EXTENDED_ERROR_ASN:
      return LDAP_SSL_EXTENDED_ERROR_ASN_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR:
      return LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR_STR;
   case LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR:
      return LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR:
      return LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT:
      return LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL_STR;
   case LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD:
      return LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED:
      return LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB:
      return LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB_STR;
   case LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR:
      return LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR_STR;
   case LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR:
      return LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR_STR;
   case LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR:
      return LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_INDEX:
      return LDAP_SSL_EXTENDED_ERROR_BAD_INDEX_STR;
   case LDAP_SSL_EXTENDED_CLOSE_FAILED:
      return LDAP_SSL_EXTENDED_CLOSE_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_DATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_DATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS:
      return LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_IO:
      return LDAP_SSL_EXTENDED_ERROR_IO_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MAC:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MAC_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_PEER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_PEER_STR;
   case LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED:
      return LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED_STR;
   case LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED:
      return LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED:
      return LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY:
      return LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER:
      return LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER_STR;
   case LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE:
      return LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE_STR;
   case LDAP_SSL_EXTENDED_WOULD_BLOCK:
      return LDAP_SSL_EXTENDED_WOULD_BLOCK_STR;
   case LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3:
      return LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3_STR;
   case LDAP_SSL_EXTENDED_MISC_INVALID_ID:
      return LDAP_SSL_EXTENDED_MISC_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE_STR;
   case LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING:
      return LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING_STR;
   default:
      return "Unknown SSL error";
   }
}


/*
** Read a record from the file.
*/
char *read_one_record(FILE *fp) {
   int len;
   char *buf, line[ LDAPMOD_MAXLINE ];
   int lcur, lmax;

   lcur = lmax = 0;
   buf = NULL;
   while (( fgets( line, sizeof(line), fp ) != NULL )) {
      /* skip comment */
      if (*line == '#')
         continue;

      else if ( *line != '\n' ) {
         len = strlen( line );
         if ( lcur + len + 1 > lmax ) {
            lmax = LDAPMOD_MAXLINE
                   * (( lcur + len + 1 ) / LDAPMOD_MAXLINE + 1 );
            if (( buf = (char *)safe_realloc( buf, lmax )) == NULL ) {
               perror( "safe_realloc" );
               exit( 1 );
            }
         }
         strcpy( buf + lcur, line );
         lcur += len;
      } else {
         if ( buf == NULL )
            continue; /* 1st line keep going */
         else
            break;
      }
   }
   return buf;
}

/*
** Read binary data from a file.
*/
int fromfile(char *path, struct berval *bv) {
   FILE * fp;
   long rlen;
   int eof;

   /* "r" changed to "rb", defect 39803. */
   if (( fp = fopen( path, "rb" )) == NULL ) {
      perror( path );
      return ( -1 );
   }

   if ( fseek( fp, 0L, SEEK_END ) != 0 ) {
      perror( path );
      fclose( fp );
      return ( -1 );
   }

   bv->bv_len = ftell( fp );

   if (( bv->bv_val = (char *)malloc( bv->bv_len )) == NULL ) {
      perror( "malloc" );
      fclose( fp );
      return ( -1 );
   }

   if ( fseek( fp, 0L, SEEK_SET ) != 0 ) {
      perror( path );
      fclose( fp );
      return ( -1 );
   }

   rlen = fread( bv->bv_val, 1, bv->bv_len, fp );
   eof = feof( fp );
   fclose( fp );

   if ( rlen != bv->bv_len ) {
      perror( path );
      free( bv->bv_val );
      bv->bv_val = NULL;
      return ( -1 );
   }

   return ( bv->bv_len );
}

/*
** Read binary data from a file specified with a URL.
*/
int fromfile_url(char *value, struct berval *bv) {
   char *file, *src, *dst;
   FILE * fp;

   if ( value == NULL || strncmp(value, "file://", 7) ) {
      fprintf(stderr, "fromfile_url: %s", ldap_err2string(LDAP_INVALID_SYNTAX));
      return -1;
   }

   file = value+7; /* skip "file://" */
   if ( strncmp(file, "localhost/", 10) == 0 ) /* skip "file://localhost/" */
      file += 10;
   else if ( *file == '/' ) /* skip "file:///" */
      file++;

   /* unescape characters */
   for (dst = src = file; (*src != '\0'); ++dst) {
      *dst = *src;
      if (*src++ != '%')
         continue;
      if ((*src >= '0') && (*src <= '9'))
         *dst = (*src++ - '0') << 4;
      else if ((*src >= 'a') && (*src <= 'f'))
         *dst = (*src++ - 'a' + 10) << 4;
      else if ((*src >= 'A') && (*src <= 'F'))
         *dst = (*src++ - 'A' + 10) << 4;
      else
         return -1;
      if ((*src >= '0') && (*src <= '9'))
         *dst += (*src++ - '0');
      else if ((*src >= 'a') && (*src <= 'f'))
         *dst += (*src++ - 'a' + 10);
      else if ((*src >= 'A') && (*src <= 'F'))
         *dst += (*src++ - 'A' + 10);
      else
         return -1;
   }
   *dst = '\0';

   /* Technically, the file path is everything after "file://[localhost]/"
   ** but there is a defacto standard which allows that path to be absolute
   ** even if it does not start w/ "/". So if the relative path is not found,
   ** try interpreting it as absolute instead.
   */

   /* Using the same method to open the file as the fromfile() function, below */
   if (( fp = fopen( file, "rb" )) == NULL ) {
      file--; /* add a leading '/' */
      if (( fp = fopen( file, "rb" )) == NULL ) {
         file++;  /* remove the added '/' so fromfile() can handle the error w/ the correct path */
      }
   }
   
   if (fp != NULL) 
      fclose(fp);

   return fromfile(file, bv);
}

/*
** Add operation to the modify structure.
*/
void addmodifyop(LDAPMod ***pmodsp, int modop, char *attr,
                 char *value, int vlen, int isURL, int isBase64) {
   LDAPMod * *pmods;
   int i, j;
   struct berval *bvp;
   int orig_modop = modop;

   /*
   ** Data can be treated as binary (wire ready) if one of the
   ** following applies:
   **     1) it was base64 encode
   **     2) charset is not defined
   **     3) read from an external file
   */
   if (isBase64 || (charset == NULL) || isURL
         || ((value != NULL) && valsfromfiles && (*value == '/'))) {
      modop |= LDAP_MOD_BVALUES;
   }

   i = 0;
   pmods = *pmodsp;
   if ( pmods != NULL ) {
      for ( ; pmods[ i ] != NULL; ++i ) {
         if ( strcasecmp( pmods[ i ]->mod_type, attr ) == 0 &&
               pmods[ i ]->mod_op == modop ) {
            break;
         }
      }
   }

   if ( pmods == NULL || pmods[ i ] == NULL ) {
      if (( pmods = (LDAPMod * *)safe_realloc( pmods, (i + 2) *
                    sizeof( LDAPMod * ))) == NULL ) {
         perror( "safe_realloc" );
         exit( 1 );
      }
      *pmodsp = pmods;
      pmods[ i + 1 ] = NULL;
      if (( pmods[ i ] = (LDAPMod * )calloc( 1, sizeof( LDAPMod ))) == NULL ) {
         perror( "calloc" );
         exit( 1 );
      }
      pmods[ i ]->mod_op = modop;
      if (( pmods[ i ]->mod_type = strdup( attr )) == NULL ) {
         perror( "strdup" );
         exit( 1 );
      }
   }

   if (value == NULL && orig_modop != LDAP_MOD_ADD)
      return ;
   if (modop & LDAP_MOD_BVALUES) {
      j = 0;
      if ( pmods[ i ]->mod_bvalues != NULL ) {
         for ( ; pmods[ i ]->mod_bvalues[ j ] != NULL; ++j ) {
            ;
         }
      }
      if (( pmods[ i ]->mod_bvalues =
               (struct berval **)safe_realloc( pmods[ i ]->mod_bvalues,
                                               (j + 2) * sizeof( struct berval *))) == NULL ) {
         perror( "safe_realloc" );
         exit( 1 );
      }
      pmods[ i ]->mod_bvalues[ j + 1 ] = NULL;
      if (( bvp = (struct berval *)malloc( sizeof( struct berval )))
            == NULL ) {
         perror( "malloc" );
         exit( 1 );
      }
      pmods[ i ]->mod_bvalues[ j ] = bvp;

      /* get value from file */
      if ( valsfromfiles && ((value != NULL) && *value == '/') ) {
         if (fromfile( value, bvp ) < 0 )
            exit(1);
      } else if (isURL) {
         if (fromfile_url(value, bvp) < 0) {
            if (!contoper)
               exit(1);
            else {
               badFile = 1;
               return ;
            }
         }
      } else {
         bvp->bv_len = vlen;
         if (( bvp->bv_val = (char *)malloc( vlen + 1 )) == NULL ) {
            perror( "malloc" );
            exit( 1 );
         }
         if (value != NULL)
            memmove( bvp->bv_val, value, vlen );
         bvp->bv_val[ vlen ] = '\0';
      }
   } else {
      j = 0;
      if ( pmods[ i ]->mod_values != NULL ) {
         for ( ; pmods[ i ]->mod_values[ j ] != NULL; ++j ) {
            ;
         }
      }
      if (( pmods[ i ]->mod_values =
               (char **)safe_realloc( pmods[ i ]->mod_values,
                                      (j + 2) * sizeof( char *))) == NULL ) {
         perror( "safe_realloc" );
         exit( 1 );
      }
      pmods[ i ]->mod_values[ j + 1 ] = NULL;
      if (value && (vlen != 0)) {
         if (( pmods[ i ]->mod_values[ j ] = strdup( value )) == NULL) {
            perror( "strdup" );
            exit( 1 );
         }
      } else {
         pmods[ i ]->mod_values[ j ] = (char *)malloc( 1 );
         *(pmods[ i ]->mod_values[ j ]) = '\0';
      }
   }

}

/*
** Print the message to request file
** If error writing to reject file, stop processing and exit
*/
void printToFile(FILE *errfp, const char *message, ...) {
   va_list ap;

   int rc = LDAP_SUCCESS;

   va_start(ap, message);
   rc = vfprintf(errfp, message, ap);
   if ((rc < 0) && (reject == 1)) {
      fprintf(stderr, "Unable to write to reject file.\n");
      exit (1);
   }
   va_end(ap);
   rc = fflush(errfp);
   if ((rc != 0) && (reject == 1)) {
      fprintf(stderr, "Unable to write to reject file.\n");
      exit (1);
   }
}

/*
** Check errors and controls, print errors when needed to requested FILE
*/
void checkErrors (FILE *errfp, int index, char *errmsg, const char *errormsg, const char *warnmsgctrl, const char *errormsgctrl, int controlres) {
   const char *action[] = {
                             T_DELETECTSTR, T_ADDCTSTR, T_MODIFYCTSTR, T_MODRDNCTSTR
                          };
   const char *option[] = {
                             T_DELETECTSTR, T_ADDCTSTR, T_MODIFYCTSTR, "rename"
                          };

   if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
      printToFile(errfp, "ldap_%s:  ", action[index]);
      if (errormsg != NULL) {
         printToFile(errfp, errormsg);
         printToFile(errfp, " --- ");
      }
      if (errormsgctrl != NULL) {
         printToFile(errfp, errormsgctrl);
         printToFile(errfp, "\n");
      }
      if (warnmsgctrl != NULL) {
         printToFile(errfp, warnmsgctrl, controlres);
         printToFile(errfp, "\n");
      }
   } else if (errmsg != NULL || errormsg != NULL) {
      if (errormsg != NULL) {
         if (*errormsg != '\0') {
            printToFile(errfp, "ldap_%s: %s\n", action[index], errormsg);
         }
      }
      if (errmsg != NULL) {
         if (*errmsg != '\0') {
            printToFile(errfp, "ldap_%s: additional info: %s\n", action[index], errmsg);
         }
      }
   } else if (verbose)
      printf("%s complete\n", option[index]);
   putchar('\n');
}

/*
** Parse results and responses, set error messages
** If reject is set 1, write errors to reject file as well as stderr
*/
int parseAndPrint (LDAP *ld, int index, int msgid, int rc2) {
   LDAPControl **returnedControls = NULL;
   LDAPMessage *res;
   char *errmsg = NULL;
   int rc;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   int errcode = 0;
   int error = 0;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;

   rc = rc2;
   if (rc == LDAP_SUCCESS) {
      error = ldap_result (ld, msgid, 1, (struct timeval *) NULL, &res );
      rc = ldap_parse_result ( ld, res, &errcode, NULL, &errmsg, NULL, &returnedControls, 1);
   }
   if (rc == LDAP_SUCCESS) {
      if (returnedControls != NULL) {
         ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
      }
   }

   if (rc2 != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc2);
      rc = rc2;
   } else if (rc != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc);
   } else if (errcode != LDAP_SUCCESS) {
      errormsg = ldap_err2string(errcode);
      rc = errcode;
   }

   if (rc == LDAP_SERVER_DOWN && reject == 0) {
      exit(rc);
   }

   if (controlerr) {
      errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
   }

   if (controlwarn) {
      warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
   }

   checkErrors(stderr, index, errmsg, errormsg, warnmsgctrl, errormsgctrl, controlres);

   if (reject == 1) {
      checkErrors(rfp, index, errmsg, errormsg, warnmsgctrl, errormsgctrl, controlres);
   }

   return rc;
}

/*
** Delete record.
*/
int dodelete(LDAP *ld, char *dn ) {
   int rc, rc2;
   int msgid = 0;

   printf( "%sdeleting entry %s\n", (!doit) ? "!" : "", dn );

   if (!doit)
      return LDAP_SUCCESS;

   rc = ldap_delete_ext( ld, dn, serverControls,
                         NULL, &msgid);
   rc2 = rc;

   rc = parseAndPrint(ld, 0, msgid, rc2);

   return rc;
}

/*
** Copy or move an entry.
*/
int domodrdn( LDAP *ld, char *dn, char *newrdn, int deleteoldrdn, char *newSuperior ) {
   int rc, rc2;
   int msgid = 0;

   printf( "%s%s %s to %s%s%s\n", ((!doit) ? "!" : ""),
           ((deleteoldrdn) ? "moving" : "copying"), dn, newrdn,
           ((newSuperior) ? "," : ""), ((newSuperior) ? newSuperior : ""));
   if (!doit)
      return LDAP_SUCCESS;

   rc = ldap_rename( ld, dn, newrdn, newSuperior, deleteoldrdn,
                     serverControls,
                     NULL, &msgid );

   rc2 = rc;

   rc = parseAndPrint(ld, 3, msgid, rc2);

   return rc;
}

/*
** Print a binary value. If charset is not specified then check to
** see if string is printable anyway.
*/
void print_binary(struct berval *bval) {
   int i, binary;

   printf( "\tBINARY (%ld bytes) ", bval->bv_len);
   if (charset == NULL) {
      binary = 0;
      for (i = 0; (i < bval->bv_len) && (!binary); ++i)
         if (!isprint(bval->bv_val[i]))
            binary = 1;
      if (!binary)
         for (i = 0; i < bval->bv_len; ++i)
            putchar(bval->bv_val[i]);
   }
   putchar('\n');
}



/*
** Modify or add an entry.
*/
int domodify( LDAP *ld, char *dn, LDAPMod **pmods, int newentry ) {
   int i, j, op, rc, rc2;
   struct berval *bvp;
   int errcode = 0;
   int error = 0;
   int msgid = 0;
   int operType = 0;

   if ( pmods == NULL ) {
      fprintf( stderr, "%s: no attributes to change or add (entry %s)\n",
               prog, dn );
      if (reject == 1) {
         printToFile(rfp, "%s: no attributes to change or add (entry %s)\n",
                     prog, dn );
      }
      return ( LDAP_PARAM_ERROR );
   }

   if ( verbose ) {
      for ( i = 0; pmods[ i ] != NULL; ++i ) {
         op = pmods[ i ]->mod_op & ~LDAP_MOD_BVALUES;
         printf( "%s %s:\n", op == LDAP_MOD_REPLACE ?
                 "replace" : op == LDAP_MOD_ADD ?
                 "add" : "delete", pmods[ i ]->mod_type );
         if (pmods[i]->mod_op & LDAP_MOD_BVALUES) {
            if (pmods[ i ]->mod_bvalues != NULL) {
               for (j = 0; pmods[i]->mod_bvalues[j] != NULL; ++j)
                  print_binary(pmods[i]->mod_bvalues[j]);
            }
         } else {
            if (pmods[i]->mod_values != NULL) {
               for (j = 0; pmods[i]->mod_values[j] != NULL; ++j)
                  printf("\t%s\n", pmods[i]->mod_values[j]);
            }
         }
      }
   }
   if ( newentry )
      printf( "%sadding new entry %s\n", (!doit) ? "!" : "", dn );
   else
      printf( "%smodifying entry %s\n", (!doit) ? "!" : "", dn );
   if (!doit)
      return LDAP_SUCCESS;

   if ( newentry )
      rc = ldap_add_ext( ld, dn, pmods, serverControls, NULL, &msgid);
   else
      rc = ldap_modify_ext( ld, dn, pmods, serverControls, NULL, &msgid);

   rc2 = rc;

   operType = newentry ? 1 : 2;

   rc = parseAndPrint( ld, operType, msgid, rc2);

   return rc;
}


int checkCryptoConsistency(LDAP * ld, char * pseudoCryptoSync, char * pseudoCryptoSalt,
                           int * isConsistent) {

   const char * attrs[] = {
                             CRYPTOSYNC_ATTR_NAME,
                             CRYPTOSALT_ATTR_NAME,
                             NULL
                          };
   char * srvCryptoSync = NULL;
   char * srvCryptoSalt = NULL;
   LDAPMessage * result = NULL;
   LDAPMessage * entry = NULL;
   BerElement * berel = NULL;
   int rc = LDAP_SUCCESS;
   char ** values = NULL;
   char* attr = NULL;


   if ((!ld) || (!pseudoCryptoSync) || (!pseudoCryptoSync) || (!isConsistent)) {
      rc = LDAP_OPERATIONS_ERROR;
      fprintf(stderr, "%s: checkCryptoConsistency() invalid parameter.\n", prog);
      goto unwind1;
   }

   *isConsistent = 0;

   rc = ldap_search_st(ld, "CN=CRYPTO,CN=LOCALHOST", LDAP_SCOPE_BASE, "OBJECTCLASS=*",
                       (char **)attrs, 0, NULL, &result);
   if (rc != LDAP_SUCCESS) {
      fprintf(stderr, "%s: ldap_search_st() failed.\n", prog);
      goto unwind1;
   }

   entry = ldap_first_entry(ld, result);
   if (entry == NULL) {
      fprintf(stderr, "%s: ldap_first_entry() failed.\n", prog);
      goto unwind1;
   }

   attr = ldap_first_attribute(ld, entry, &berel);
   if (attr == NULL) {
      fprintf(stderr, "%s: ldap_first_attribute() failed.\n", prog);
      goto unwind1;

   }

   while (attr) {

      if (strcasecmp(attr, CRYPTOSYNC_ATTR_NAME) == 0) {
         values = ldap_get_values(ld, entry, attr);
         if (values) {
            srvCryptoSync = strdup(values[0]);
            if (srvCryptoSync == NULL) {
               rc = LDAP_NO_MEMORY;
               fprintf(stderr, "%s: strdup(CRYPTOSYNC_ATTR_NAME) failed.\n", prog);
               goto unwind1;
            }
            ldap_value_free(values);
         }

      } else if (strcasecmp(attr, CRYPTOSALT_ATTR_NAME) == 0) {
         values = ldap_get_values(ld, entry, attr);
         if (values) {
            srvCryptoSalt = strdup(values[0]);
            if (srvCryptoSalt == NULL) {
               rc = LDAP_NO_MEMORY;
               fprintf(stderr, "%s: strdup(CRYPTOSALT_ATTR_NAME) failed.\n", prog);
               goto unwind1;
            }
            ldap_value_free(values);
         }
      }

      attr = ldap_next_attribute(ld, entry, berel);

   } /* end while(attr) */

   if (!strcmp(srvCryptoSync, pseudoCryptoSync) &&
         !strcmp(srvCryptoSync, pseudoCryptoSync)) {
      *isConsistent = 1;
   }

unwind1:

   if (berel != NULL) {
      ldap_memfree((char*) berel);
   }
   if (result) {
      ldap_msgfree(result);
   }
   if (srvCryptoSync) {
      free (srvCryptoSync);
   }
   if (srvCryptoSalt) {
      free(srvCryptoSalt);
   }

   return (rc);
}


/*
** Process an ldif record.
*/
int process_ldif_rec(LDAP **ld, char *rbuf) {
   char *line, *dn, *type, *value, *newrdn, *p, *newSuperior;
   char *mod_attr_type; /* 92899 */
   int is_url, is_b64;
   int rc, linenum, vlen, modop;
   int expect_modop, expect_sep, expect_ct, expect_newrdn;
   int expect_moddn_args, deleteoldrdn;
   int saw_replica, use_record, new_entry, delete_entry, got_all;
   LDAPMod **pmods;
   int version = 0;
   int str_rc = 0;
   LDAPControl** ldifControls = NULL;
   int control_count = 0;
   static int recNumber = 0;
   int isConsistent = 0;
   
   badFile = 0;
   new_entry = (operation == LDAPMODIFY_ADD);
   rc = got_all = saw_replica = delete_entry = expect_modop = 0;
   expect_moddn_args = expect_newrdn = expect_sep = expect_ct = 0;
   linenum = 0;
   deleteoldrdn = 1;
   pmods = NULL;
   dn = newrdn = newSuperior = NULL;
   modop = 0;
   type = NULL;
   value = NULL;
   mod_attr_type = NULL;/* 92899 */

   while ( rc == 0 && ( line = str_getline( &rbuf )) != NULL ) {
      ++linenum;

      /* Is this a separator line ("-")? */
      if ( expect_sep && strcasecmp( line, T_MODSEPSTR ) == 0 ) {
         /*
         ** If modifier has not been added yet then go ahead and add
         ** it. The can happen on sequences where there are no
         ** attribute values, such as:
         ** DELETE: title
         ** -
         */
         if (value != NULL)
            addmodifyop(&pmods, modop, value, NULL, 0, 0, 0);
         value = NULL;
         expect_sep = 0;
         expect_modop = 1;
         mod_attr_type = NULL; /* 92899 */
         continue;
      }

      str_rc = str_parse_line_v_or_bv(line, &type, &value, &vlen, ignoreSpaces, &is_url, &is_b64);
      if (str_rc < 0 ) {
         if (str_rc != -2) /* special case, indicates NULL attribute value, we want to modify an attrib to nul val */
         {
            fprintf( stderr, "%s: invalid format (line %d of entry: %s)\n",
                     prog, linenum, dn == NULL ? "" : dn );
            if (reject == 1) {
               printToFile( rfp, "%s: invalid format (line %d of entry: %s)\n",
                            prog, linenum, dn == NULL ? "" : dn );
            }
            rc = LDAP_PARAM_ERROR;
            break;
         }
      }

      if ((linenum == 1) && (strcmp(type, "version") == 0)) {
         version = atoi(value);
         continue;
      }

      if ((linenum == 2) && (version == 1) &&
            (strcmp(type, "charset") == 0)) {
         if (charset != NULL)
            free(charset);
         if (value == NULL) {
            fprintf( stderr, "%s: invalid format (line %d of entry: %s)\n",
                     prog, linenum, dn == NULL ? "" : dn );
            if (reject == 1) {
               printToFile( rfp, "%s: invalid format (line %d of entry: %s)\n",
                            prog, linenum, dn == NULL ? "" : dn );
            }
            rc = LDAP_PARAM_ERROR;
            break;
         }
         charset = strdup(value);
         if ((rc = ldap_set_iconv_local_charset(charset)) != LDAP_SUCCESS) {
            fprintf(stderr, "unsupported charset %s\n", charset);
            if (reject == 1) {
               printToFile(rfp, "unsupported charset %s\n", charset);
            }
            break;
         }
         ldap_set_option(*ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
         continue;
      }

      if ( dn == NULL ) {
         /*The value cannot be null at this point*/
         if (value == NULL) {
            fprintf( stderr, "%s: invalid format (line %d of entry: %s)\n",
                     prog, linenum, dn == NULL ? "" : dn );
            if (reject == 1) {
               printToFile( rfp, "%s: invalid format (line %d of entry: %s)\n",
                            prog, linenum, dn == NULL ? "" : dn );
            }
            rc = LDAP_PARAM_ERROR;
            break;
         }
         if ( strcasecmp( type, T_DN_STR ) == 0 ) {
            if (( dn = strdup( value )) == NULL ) {
               perror( "strdup" );
               exit( 1 );
            }
            expect_ct = 1;

            if ((pseudoCryptoSyncEntry == 0) && (!strcasecmp(dn, "cn=crypto,cn=localhost"))) {
               pseudoCryptoSyncEntry = 1;
            }
         } else { /* expecting dn: <...> */
            fprintf( stderr, "%s: invalid format (line %d of entry: %s)\n",
                     prog, linenum, dn == NULL ? "" : dn );
            if (reject == 1) {
               printToFile( rfp, "%s: invalid format (line %d of entry: %s)\n",
                            prog, linenum, dn == NULL ? "" : dn );
            }
            rc = LDAP_PARAM_ERROR;
            break;
         }
         continue; /* skip all lines until we see "dn:" */
      }


      /* if any controls are set in an ldif file that
      * is working with a <V3 Server. The controls will be
      * not be passed to the server. however they will be parsed.
      */
      if ( strcmp(type, T_CONTROLSTR) == 0 ) {
         int base64 = 0;
         int check_url = 0 ;
         LDAPControl* newControl = NULL;
         if (control_count == 0) {
            if ((rc = ldap_copy_controls(&ldifControls, opControls)) != LDAP_SUCCESS) {
               fprintf( stderr, "ldap_modify: ldap_copy_controls: %d, %s\n",
                        rc, ldap_err2string( rc ) );
               if (reject == 1) {
                  printToFile( rfp, "ldap_modify: ldap_copy_controls: %d, %s\n",
                               rc, ldap_err2string( rc ) );
               }
               exit(rc);
            }
         }
         control_count++;
         newControl = (LDAPControl* )string2control(value, &base64, &check_url);
         if (newControl)
            ldap_insert_control(newControl, &ldifControls);
         else {
            fprintf(stderr, "Failed to create control.\n");
            if (reject == 1) {
               printToFile(rfp, "Failed to create control.\n");
            }
            exit (1);
         }
         continue;
      }


      if ( expect_ct ) {
         expect_ct = 0;

         /* this is an ldif-change-record */
         if ( strcasecmp( type, T_CHANGETYPESTR ) == 0 ) {
            /*The value cannot be null at this point*/
            if (value == NULL) {
               fprintf( stderr, "%s: invalid format (line %d of entry: %s)\n",
                        prog, linenum, dn == NULL ? "" : dn );
               if (reject == 1) {
                  printToFile( rfp, "%s: invalid format (line %d of entry: %s)\n",
                               prog, linenum, dn == NULL ? "" : dn );
               }
               rc = LDAP_PARAM_ERROR;
               break;
            }
            if ( strcasecmp( value, T_MODIFYCTSTR ) == 0 ) {
               new_entry = 0;
               expect_modop = 1;
            } else if ( strcasecmp( value, T_ADDCTSTR ) == 0 ) {
               modop = LDAP_MOD_ADD;
               new_entry = 1;
            } else if ( (strcasecmp( value, T_MODRDNCTSTR ) == 0) ||
                        (strcasecmp( value, T_MODDNCTSTR ) == 0) ) {
               expect_newrdn = 1;
            } else if ( strcasecmp( value, T_DELETECTSTR ) == 0 ) {
               got_all = delete_entry = 1;
            } else {
               fprintf( stderr,
                        "%s:  unknown %s \"%s\" (line %d of entry: %s)\n",
                        prog, T_CHANGETYPESTR, value, linenum, dn );
               if (reject == 1) {
                  printToFile( rfp,
                               "%s:  unknown %s \"%s\" (line %d of entry: %s)\n",
                               prog, T_CHANGETYPESTR, value, linenum, dn );
               }
               rc = LDAP_PARAM_ERROR;
            }
            continue;

            /* this is an ldif-attrval-record */
         } else {
            if (operation == LDAPMODIFY_ADD) {
               new_entry = 1;
               modop = LDAP_MOD_ADD;
            } else
               modop = LDAP_MOD_REPLACE;
         }
      }

      if (expect_modop) {
         expect_modop = 0;
         expect_sep = 1;
         /* 92899 - attribute type is required after mod-op */
         if ( value && strlen( value ) )
            mod_attr_type = value;
         else {
            fprintf(stderr,
                    "%s: attribute type expected after \"%s\" (line %d of entry: %s)\n",
                    prog, type, linenum, dn);
            rc = LDAP_PARAM_ERROR;
            continue;
         }
         if ( strcasecmp( type, T_MODOPADDSTR ) == 0 ) {
            modop = LDAP_MOD_ADD;
            continue;
         } else if ( strcasecmp( type, T_MODOPREPLACESTR ) == 0 ) {
            modop = LDAP_MOD_REPLACE;
            continue;
         } else if ( strcasecmp( type, T_MODOPDELETESTR ) == 0 ) {
            modop = LDAP_MOD_DELETE;
            continue;
         } else {
            fprintf(stderr,
                    "%s:  unknown mod_spec \"%s\" (line %d of entry: %s)\n",
                    prog, type, linenum, dn);
            if (reject == 1) {
               printToFile(rfp,
                           "%s:  unknown mod_spec \"%s\" (line %d of entry: %s)\n",
                           prog, type, linenum, dn);
            }
            rc = LDAP_PARAM_ERROR;
            continue;
         }
      }

      if ( expect_newrdn ) {
         if ( strcasecmp( type, T_NEWRDNSTR ) == 0 ) {
            /*The value cannot be null at this point*/
            if (value == NULL) {
               fprintf( stderr, "%s: invalid format (line %d of entry: %s)\n",
                        prog, linenum, dn == NULL ? "" : dn );
               if (reject == 1) {
                  printToFile( rfp, "%s: invalid format (line %d of entry: %s)\n",
                               prog, linenum, dn == NULL ? "" : dn );
               }
               rc = LDAP_PARAM_ERROR;
               break;
            }
            if (( newrdn = strdup( value )) == NULL ) {
               perror( "strdup" );
               exit( 1 );
            }
            expect_moddn_args = 1;
            expect_newrdn = 0;
         } else {
            fprintf( stderr, "%s: expecting \"%s:\" but saw \"%s:\" (line %d of entry %s)\n",
                     prog, T_NEWRDNSTR, type, linenum, dn );
            if (reject == 1) {
               printToFile( rfp, "%s: expecting \"%s:\" but saw \"%s:\" (line %d of entry %s)\n",
                            prog, T_NEWRDNSTR, type, linenum, dn );
            }
            rc = LDAP_PARAM_ERROR;
         }
      } else if ( expect_moddn_args ) {
         if ( strcasecmp( type, T_DELETEOLDRDNSTR ) == 0 ) {
            deleteoldrdn = ( *value == '0' ) ? 0 : 1;
         } else if ( strcasecmp( type, T_NEWSUPERIORSTR ) == 0 ) {
            if ( newSuperior == NULL ) {
               if (( newSuperior = strdup( value )) == NULL ) {
                  perror( "strdup" );
                  exit( 1 );
               }
            } else {
               fprintf( stderr, "%s: invalid format (line %d of entry %s)\n",
                        prog, linenum, dn );
               if (reject == 1) {
                  printToFile( rfp, "%s: invalid format (line %d of entry %s)\n",
                               prog, linenum, dn );
               }
               rc = LDAP_PARAM_ERROR;
            }
         } else {
            fprintf( stderr, "%s: expecting \"%s:\" or \"%s:\" but saw \"%s:\" (line %d of entry %s)\n",
                     prog, T_DELETEOLDRDNSTR, T_NEWSUPERIORSTR, type, linenum, dn );
            if (reject == 1) {
               printToFile( rfp, "%s: expecting \"%s:\" or \"%s:\" but saw \"%s:\" (line %d of entry %s)\n",
                            prog, T_DELETEOLDRDNSTR, T_NEWSUPERIORSTR, type, linenum, dn );
            }
            rc = LDAP_PARAM_ERROR;
         }
      } else if ( got_all ) {
         fprintf( stderr, "%s: extra lines at end (line %d of entry %s)\n",
                  prog, linenum, dn );
         if (reject == 1) {
            printToFile( rfp, "%s: extra lines at end (line %d of entry %s)\n",
                         prog, linenum, dn );
         }
         rc = LDAP_PARAM_ERROR;
      } else {

         if (pseudoCryptoSyncEntry == 1) {
            if (!strcasecmp(type, "ibm-slapdCryptoSync")) {
               pseudoCryptoSync = strdup(value);
            }
            if (!strcasecmp(type, "ibm-slapdCryptoSalt")) {
               pseudoCryptoSalt = strdup(value);
            }
         }

         /*
         ** 92899 Verify the attribute after change type matches the attribute being modified
         */
         if ( mod_attr_type == NULL || strcasecmp(mod_attr_type, type) == 0 ) {
            addmodifyop(&pmods, modop, type, value, vlen, is_url, is_b64);
         } else {
            fprintf(stderr,
                    "%s: expected \"%s\" but found \"%s\" (line %d of entry: %s)\n",
                    prog, mod_attr_type, type, linenum, dn);
            rc = LDAP_PARAM_ERROR;
         }

         type = NULL;
         value = NULL;
         if (badFile)
            return (LDAP_PARAM_ERROR);
      }
   }

   /*
   ** If last separator is missing go ahead and handle it anyway, even
   ** though it is technically invalid ldif format.
   */
   if (expect_sep && (value != NULL))
      addmodifyop(&pmods, modop, value, NULL, 0, 0, 0);

   /*set controls the controls*/
   if (ldapversion == LDAP_VERSION3) {
      if ( ldifControls )
         serverControls = ldifControls;
      else
         serverControls = opControls;
   }

   if ( rc == 0 ) {
      if (delete_entry) {
         rc = dodelete(*ld, dn);
      } else if (newrdn != NULL) {
         rc = domodrdn(*ld, dn, newrdn, deleteoldrdn, newSuperior);
      } else if ((dn != NULL) && (pseudoCryptoSyncEntry == 0)) {
         rc = domodify(*ld, dn, pmods, new_entry );
      }

      recNumber++;

      if (pseudoCryptoSyncEntry == 1) {

         /* The pseudo crypto sync entry should be in the LDIF header */

         if (recNumber > 3) {
            fprintf(stderr, "%s: crypto synch entry outside of LDIF header section\n", prog);
            exit(1);
         }

         /* Reset pseudo crypto sync entry indicator flag */

         pseudoCryptoSyncEntry = 0;

         if (pseudoCryptoSync && pseudoCryptoSalt) {
            rc = checkCryptoConsistency(*ld, pseudoCryptoSync, pseudoCryptoSalt, &isConsistent);
            if (rc != LDAP_SUCCESS) {
               fprintf(stderr, "%s: checkCryptoConsistency() failed\n", prog);
               exit(1);
            }
            if (isConsistent == 0) {
               fprintf(stderr, "%s: cryptographic inconsistency with target server\n", prog);
               exit(1);
            }
         } else {
            fprintf(stderr, "%s: invalid crypto synch entry\n", prog);
            rc = LDAP_PARAM_ERROR;
            exit(1);
         }

         /* Unbind and rebind with the AES crypto control */

         if (tls) {
            ldap_stop_tls_s_np(*ld);
         } else {
            ldap_unbind(*ld);
         }
         
         *ld = NULL;

         rc = ldap_add_control(LDAP_AES_BIND_CONTROL_OID, 0, NULL, LDAP_OPT_ON, &bindControls);
         if (rc != LDAP_SUCCESS) {
            fprintf(stderr, "%s: ldap_add_control() failed\n", prog);
            exit(1);
         }

         *ld = connect_to_server();
         if (!(*ld)) {
            fprintf(stderr, "%s: rebind call to connect_to_server() failed\n", prog);
            exit(1);
         }

      } /* end if(pseudoCryptoSyncEntry) */

   }

   if ( ldifControls )
      ldap_controls_free(ldifControls);
   if (dn != NULL)
      free( dn );
   if ( newrdn != NULL )
      free( newrdn );
   if ( pmods != NULL )
      ldap_mods_free( pmods, 1 );
   if ( newSuperior != NULL )
      free( newSuperior );

   return ( rc );
}

/*
** Process a mod record.
*/
int process_ldapmod_rec( LDAP *ld, char *rbuf ) {
   char *line, *dn, *p, *q, *attr, *value;
   int rc, linenum, modop, charcount;
   LDAPMod * *pmods;

   pmods = NULL;
   dn = NULL;
   linenum = 0;
   line = rbuf;
   rc = 0;

   while ( rc == 0 && rbuf != NULL && *rbuf != '\0' ) {
      ++linenum;
      if (( p = strchr( rbuf, '\n' )) == NULL ) {
         rbuf = NULL;
      } else {
         if ( *(p - 1) == '\\' ) { /* lines ending in '\' are continued */
            charcount = strlen(p);
            memmove( p - 1, p, charcount + 1 ); /* include the terminating null */
            rbuf = p;
            continue;
         }
         *p++ = '\0';
         rbuf = p;
      }

      if ( dn == NULL ) { /* first line contains DN */
         if (( dn = strdup( line )) == NULL ) {
            perror( "strdup" );
            exit( 1 );
         }
      } else {
         if (( p = strchr( line, '=' )) == NULL ) {
            value = NULL;
            p = line + strlen( line );
         } else {
            *p++ = '\0';
            value = p;
         }

         for ( attr = line; *attr != '\0' && isspace( *attr ); ++attr ) {
            ; /* skip attribute leading white space */
         }

         for ( q = p - 1; q > attr && isspace( *q ); --q ) {
            *q = '\0'; /* remove attribute trailing white space */
         }

         if ( value != NULL ) {
            while ( isspace( *value )) {
               ++value;  /* skip value leading white space */
            }
            for ( q = value + strlen( value ) - 1; q > value &&
                  isspace( *q ); --q ) {
               *q = '\0'; /* remove value trailing white space */
            }
            if ( *value == '\0' ) {
               value = NULL;
            }

         }

         if ((value == NULL) && (operation == LDAPMODIFY_ADD)) {
            fprintf( stderr, "%s: missing value on line %d (attr is %s)\n",
                     prog, linenum, attr );
            if (reject == 1) {
               printToFile( rfp, "%s: missing value on line %d (attr is %s)\n",
                            prog, linenum, attr );
            }
            rc = LDAP_PARAM_ERROR;
         } else {
            switch ( *attr ) {
            case '-':
               modop = LDAP_MOD_DELETE;
               ++attr;
               break;
            case '+':
               modop = LDAP_MOD_ADD;
               ++attr;
               break;
            default:
               modop = (operation == LDAPMODIFY_REPLACE)
                       ? LDAP_MOD_REPLACE : LDAP_MOD_ADD;
               break ;
            }

            addmodifyop( &pmods, modop, attr, value,
                         ( value == NULL ) ? 0 : strlen( value ), 0, 0);
         }
      }

      line = rbuf;
   }

   /* set controls d97205*/
   if (ldapversion == LDAP_VERSION3) {
      serverControls = opControls;
   }

   if ( rc == 0 ) {
      if ( dn == NULL )
         rc = LDAP_PARAM_ERROR;
      else
         rc = domodify(ld, dn, pmods, (operation == LDAPMODIFY_ADD));
   }

   if ( pmods != NULL )
      ldap_mods_free( pmods, 1 );
   if ( dn != NULL )
      free( dn );

   return ( rc );
}


int isAdminUserType(LDAP * ld, int * answer) {

   struct berval *reqVal;
   const char *oid = LDAP_USER_TYPE_REQ_OID;
   char *oidres = NULL;
   struct berval *valres = NULL;
   BerElement * ber = NULL;
   char * userType = NULL;
   int rc = LDAP_SUCCESS;

   *answer = 0;

   reqVal = ldap_create_user_type_request();
   if (!reqVal) {
      rc = LDAP_NO_MEMORY;
      fprintf(stderr, "%s: ldap_create_user_type_request() failed.\n", prog);
      goto unwind1;
   }

   rc = ldap_extended_operation_s(ld, oid, reqVal, opControls, NULL, &oidres, &valres);
   if (rc != LDAP_SUCCESS) {
      fprintf(stderr, "%s: ldap_extended_operation_s() failed.\n", prog);
      goto unwind1;
   }

   ber = ber_init2(valres);
   if (!ber) {
      rc = LDAP_OPERATIONS_ERROR;
      fprintf(stderr, "%s: ber_init2() failed.\n", prog);
      goto unwind1;
   }

   ber_scanf(ber, "{a", userType);
   if (!userType) {
      rc = LDAP_OPERATIONS_ERROR;
      fprintf(stderr, "%s: ber_scanf() failed.\n", prog);
      goto unwind1;
   }

   if (!strcasecmp(userType, "root_administrator") ||
         !strcasecmp(userType, "admin_group_member")) {
      *answer = 1;
   }

unwind1:

   if (reqVal) {
      ber_bvfree(reqVal);
   }
   if (oidres) {
      ldap_memfree(oidres);
   }
   if (valres) {
      ber_bvfree(valres);
   }

   return (rc);

}


int main( int argc, char **argv ) {
   char *rbuf, *start, *p, *q, *tmpstr, *rejentry;
   int rc, i, use_ldif, rejcount = 0;
   LDAP *ld = NULL;

   /* Strip off any path info on program name */
#if defined( WIN32 ) || defined( _WIN32 )

   if ((prog = strrchr(argv[0], '\\')) != NULL)
      ++prog;
   else
      prog = argv[0];
#else

   if (prog = strrchr(argv[0], '/'))
      ++prog;
   else
      prog = argv[0];
#endif

#if defined( WIN32 ) || defined( _WIN32 )
   /* Convert string to lowercase */
   for (i = 0; prog[i] != '\0'; ++i)
      prog[i] = tolower(prog[i]);
   /* Strip ending .exe from program name */
   if ((tmpstr = strstr(prog, ".exe")) != NULL)
      * tmpstr = '\0';
#endif

   if ( strcmp( prog, "ldapadd" ) == 0 )
      operation = LDAPMODIFY_ADD;

   /* parse command line arguments */
   parse_arguments(argc, argv);

   /*connect to server*/
   if (doit)
      ld = connect_to_server();

   if (fp != stdin)
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_OFF);

   /* do operations */
   rc = 0;
   while ((rc == 0 || contoper) && (rbuf = read_one_record( fp )) != NULL ) {
      /*
      ** we assume record is ldif/slapd.replog if the first line
      ** has a colon that appears to the left of any equal signs, OR
      ** if the first line consists entirely of digits (an entry id)
      */
      use_ldif = ( p = strchr( rbuf, ':' )) != NULL &&
                 ( q = strchr( rbuf, '\n' )) != NULL && p < q &&
                 (( q = strchr( rbuf, '=' )) == NULL || p < q );

      if (reject == 1) {
         if (( rejentry = (char *)calloc(1, strlen(rbuf) + 1)) == NULL ) {
            perror( "calloc" );
            exit( 1 );
         }
         strncpy(rejentry, rbuf, strlen(rbuf));
      }

      start = rbuf;

      if ( !use_ldif && ( q = strchr( rbuf, '\n' )) != NULL ) {
         for ( p = rbuf; p < q; ++p ) {
            if ( !isdigit( *p ))
               break;
         }
         if ( p >= q ) {
            use_ldif = 1;
            start = q + 1;
         }
      }

      if ( use_ldif ) {
         rc = process_ldif_rec(&ld, start);
      } else {
         rc = process_ldapmod_rec(ld, start );
      }

      if ((reject == 1) && (rc != LDAP_SUCCESS)) {
         /* write entry to reject file */
         printToFile(rfp, "%s\n", rejentry);
         rejcount++;

      }
      if (reject == 1)
         free(rejentry);
      free( rbuf );
   }

   if (doit) {
      if ( tls )
         ldap_stop_tls_s_np ( ld );
      else
         ldap_unbind(ld);
   }

   if (reject == 1) {
      fprintf(stdout, "%d entries were written to the reject file\n", rejcount);
      fclose(rfp);
   }

   if (bindControls)
      ldap_controls_free(bindControls);
   if (opControls)
      ldap_controls_free(opControls);
   exit(rc);
}

